"""This module stores material classes.
"""

from . import base_classes

class Material(base_classes.Idobj):
    """Makes a linear elastic meterial.

    Args:
        name (str): a unique matl name

    Attributes:
        - id (int): material id number
        - mechtype (str): defines mechanical material type: linear or non-linear material
        - density (float): density in mass/volume units
        - pratio (float): poisson's ratio (unitless)
        - youngs (float): young's modulus in Force/area = stress units
        - mechtpye (str): mechanical material type
            options: 'linear' or 'nonlinear'
        - exponent (float): exponent of Ramberg-Osgood stress-strain equation
        - yield_stress (float): yield stress of material
        - yield_offset (float): yield offset of Ramberg-Osgood stress-strain equation
        - conductivity (float): thermal conductivity, Power / (distance-temp)
        - spec_heat (float): specific heat (energy/(mass-temp)
        - thermal_exp (dict): a dict storing temperature dependent thermal
            -- expansion properties

            -- Thermal expansion is in strain per temperature

            - dict['data'] = zip of (temp, thermal_expansion)
            - dict['tzero'] = the temperature zero point
    """

    def __init__(self, name):
        self.name = name
        base_classes.Idobj.__init__(self)
        # mechanical
        self.mechtype = None
        self.density = None
        self.youngs = None
        self.pratio = None
        self.exponent = None
        self.yield_stress = None
        self.yield_offset = None
        # thermal
        self.conductivity = None
        self.spec_heat = None
        # thermal growth
        self.thermal_exp = {}

    def set_mech_props(self, density, youngs, pratio, mechtype='linear', exponent=0., yield_stress=0., yield_offset=0.):
        """Sets the mechanical properties: density, youngs, poisson_ratio.

        Args:
          - density (float): density in mass/volume units
          - pratio (float): poisson's ratio (unitless)
          - youngs (float): young's modulus in Force/area = stress units
          
        Kargs:
          - mechtpye (str): mechanical material type
            options: 'linear' or 'nonlinear'
          - exponent (float): exponent of Ramberg-Osgood stress-strain equation
          - yield_stress (float): yield stress of material
          - yield_offset (float): yield offset of Ramberg-Osgood stress-strain equation
        """
        self.density = density
        self.youngs = youngs
        self.pratio = pratio
        self.mechtype = mechtype
        self.exponent = exponent
        self.yield_stress = yield_stress
        self.yield_offset = yield_offset

    def set_therm_props(self, conductivity, spec_heat):
        """Sets the thermal properties: conductivity, specifc_heat.

        Args:
          - conductivity (float): Power / (distance-temp)
          - spec_heat (float): specific heat (energy/(mass-temp)
        """
        self.conductivity = conductivity
        self.spec_heat = spec_heat

    def set_therm_expan(self, alphas, temps=None, tzero=None):
        """Sets the thermal expansion of the material.

        Args:
          - alphas (float or list): list of thermal expansion alphas
                         length must be the same as the passed temps
          - temps (list): list of temperatures
          - tzero (float): temperature zero point
        """
        self.thermal_exp = {}
        isnum = (isinstance(alphas, float) or isinstance(alphas, int))
        if isinstance(alphas, list) and isinstance(temps, list):
            self.thermal_exp['data'] = zip(temps, alphas)
        elif temps == None and isnum:
            self.thermal_exp['data'] = [alphas]
        if tzero != None:
            self.thermal_exp['tzero'] = tzero

    def ccx(self):
        """Returns a list of text strings for ccx defining the material."""
        res = []
        res.append('*MATERIAL,NAME='+self.name)
        
        if self.mechtype == 'linear':
            res.append('*ELASTIC')
            res.append(str(self.youngs)+','+str(self.pratio))
        elif self.mechtype == 'nonlinear':
            res.append('*DEFORMATION PLASTICITY')
            res.append(','.join([str(self.youngs),str(self.pratio),str(self.yield_stress),str(self.exponent),str(self.yield_offset)]))
        if self.density != None:
            res.append('*DENSITY')
            res.append(str(self.density))
        if self.conductivity != None:
            res.append('*CONDUCTIVITY')
            res.append(str(self.conductivity))
        if self.spec_heat != None:
            res.append('*SPECIFIC HEAT')
            res.append(str(self.spec_heat))
        if self.thermal_exp != {}:
            if 'tzero' in self.thermal_exp:
                res.append('*EXPANSION,ZERO='+str(self.thermal_exp['tzero']))
            else:
                res.append('*EXPANSION')
            for pair in self.thermal_exp['data']:
                if len(pair) == 1:
                    res.append(str(pair[0]))
                else:
                    # temp, val
                    res.append('%f %f' % (pair[0], pair[1]))
        return res



"""This module stores the Part class. It is used to make 2D parts.
"""
import numpy as np # needed for linspace on hole creation

from . import base_classes
from . import geometry #point, line, area

class Part(base_classes.Idobj):
    """This makes a part.

    Args:
        parent: parent FeaModel

    Attributes:
        __fea (FeaModel): parent FeaModel
        points (list): list or part points, excludes arc centers
        allpoints (list): list or part points, includes arc centers
        lines (list): list of all Line and Arc that make the part
        signlines (list): list of all SignLine and SignArc that make the part
        __cursor (Point): location of a cursor drawing the part
        __holemode (bool): if True, lines will be added to holes, otherwise,
            they'll be added to areas
        areas (list): list of Area that make up the part
        left (list): list the parts leftmost lines, they must be vertical
        right (list): list the parts rightmost lines, they must be vertical
        top (list): list the parts top lines, they must be horizontal
        bottom (list): list the parts bottom lines, they must be horizontal
        center (Point): the area centroid of the part
        nodes (list): list of part's nodes
        elements (list): list of part's elements
    """

    def __init__(self, feamodel):
        self.fea = feamodel
        self.__cursor = geometry.Point(0, 0)
        self.areas = [] # top area is the buffer
        # make the buffer
        area = self.fea.areas.append(geometry.Area(self, []))
        self.areas.append(area)
        base_classes.Idobj.__init__(self)
        self.left = None
        self.right = None
        self.top = None
        self.bottom = None
        self.center = None
        self.nodes = []
        self.elements = []
        self.__holemode = False
        self.fea.parts.append(self)

    def __hash__(self):
        """Returns the item's id as its hash."""
        return self.id

    @property
    def lines(self):
        """Returns list of part lines."""
        lines = set()
        for area in self.areas:
            lines.update(area.lines)
        return list(lines)

    @property
    def signlines(self):
        """Returns list of part signline and signarc."""
        lines = set()
        for area in self.areas:
            lines.update(area.signlines)
        return list(lines)

    @property
    def points(self):
        """Returns list of part points, excludes arc centers."""
        points = set()
        lines = self.lines
        for line in lines:
            points.update(line.points)
        return list(points)

    @property
    def allpoints(self):
        """Returns list of part points, includes arc centers."""
        points = set()
        lines = self.lines
        for line in lines:
            points.update(line.allpoints)
        return list(points)

    def get_item(self, item):
        """"Returns the part's item(s) requested by the passed string.

        Args:
            item (str): string requesting item(s)

                * Valid examples: 'P0', 'L0', 'left', 'A0'

        Returns:
            item(s) or None: If items are found they are returned

                * If there is only one item it is returned
                * If there are multiple items, they are returned as a list
                * If no items are found None is returned
        """
        if item in ['left', 'right', 'top', 'bottom']:
            items = getattr(self, item)
            return items
        elif item[0] == 'P':
            # get point
            items = self.points
            num = int(item[1:])
            res = [a for a in items if a.id == num]
            return res[0]
        elif item[0] == 'L':
            # get line
            items = self.signlines
            num = int(item[1:])
            items = [a for a in items if a.id == num]
            return items[0]
        elif item[0] == 'A':
            # get area
            items = self.areas
            num = int(item[1:])
            items = [a for a in items if a.id == num]
            return items[0]
        else:
            print('Unknown item! Please pass the name of a point, line or area!')
            return None

    def get_name(self):
        """Returns the part name based on id number."""
        return 'PART'+str(self.id)

    def __set_side(self, side):
        """Sets the part.side to a list of lines on that side of the part.

        Used to set the part.left, part.right, part.top, part.bottom sides.

        Args:
            side (string): 'left', 'right', 'top','bottom'
        """
        # set index and axis, ind=0 is low side, ind=-1 is high side
        inds = {'left':0, 'right':-1, 'top':-1, 'bottom':0}
        axes = {'left':'y', 'right':'y', 'top':'x', 'bottom':'x'}
        ind = inds[side]
        axis = axes[side]

        # loc = 'left', ind = 0, axis = 'y'
        points = self.points
        # sort the points low to high
        points = sorted(points, key=lambda pt: getattr(pt, axis))
        # store the target value
        target_value = getattr(points[ind], axis)
        res = []
        lines = self.signlines
        for sline in lines:
            if isinstance(sline, geometry.SignLine):
                pt_axis_vals = [getattr(pt, axis) for pt in sline.points]
                pt_dist_vals = [abs(target_value - pt_axis_val) for pt_axis_val in pt_axis_vals]
                if all([pt_dist_val < geometry.ACC for pt_dist_val in pt_dist_vals]):
                    # line is on the left side
                    res.append(sline)
        setattr(self, side, res)

    def goto(self, x, y, holemode=False):
        """Moves the part cursor to a location.

        If that location has a point at it, use it.
        If not, make a new point at that location.

        Args:
            x (float): x-coordinate of the point to go to
            y (float): y-coordinate of the point to go to
            holemode (bool): if True, we start drawing a hole here, otherwise
                we start drawing an area

        Returns:
            self.__cursor (Point): returns the updated cursor point
        """
        [pnew, already_exists] = self.__make_get_pt(x, y)

        if already_exists:
            if self.areas[-1].closed == True:
                # make a new area if the old area is already closed and we're
                # going to an existing point
                area = self.fea.areas.append(geometry.Area(self, []))
                self.areas.append(area)

        # return cursor
        self.__cursor = pnew
        self.__holemode = holemode
        return self.__cursor

    def __get_point(self, point):
        """Returns point if found, None otherwise."""
        points = self.allpoints
        found_point = None
        for apoint in points:
            dist = point - apoint
            dist = dist.length()
            if dist < geometry.ACC:
                # point already exists in part, use it
                found_point = apoint
                break
        return found_point

    def __make_get_pt(self, x, y):
        """Gets a point if it exists, makes it if it doesn't. Returns the point.

        Use this when you need a point made in the part, and you want to
        use an extant point if one is available.

        Args:
            x (float): point x-coordinate
            y (float): point y-coordinate

        Returns:
            list:
                list[0]: Point
                list[1]: boolean, True = the point already existed
        """
        thept = geometry.Point(x, y)
        pfound = self.__get_point(thept)
        pexists = True
        if pfound == None:
            pfound = thept
            self.fea.register(pfound)
            pexists = False
        return [pfound, pexists]

    def __calc_area_center(self):
        """Calculates and returns the part and centroid Point.

        Returns:
            list: [area, Point]
        """
        val_list = []
        for area in self.areas:
            if area.closed == True:
                aval, cval = area.area, area.center
                val_list.append([aval, cval])
        a_sum = sum([aval[0] for aval in val_list])
        cxa_sum = sum([center.x*aval for [aval, center] in val_list])
        cya_sum = sum([center.y*aval for [aval, center] in val_list])
        cx_val = cxa_sum/a_sum
        cy_val = cya_sum/a_sum
        center = geometry.Point(cx_val, cy_val)
        return [a_sum, center]

    def __make_get_sline(self, lnew):
        """Returns a signed line or arc, makes it if it needs to.

        Args:
            lnew (Line or Arc or SignLine or SignArc): Line or Arc to make

        Returns:
            list:
                list[0]: SignLine or SignArc
                list[1]: boolean, True = the line already existed
        """
        lpos = lnew.signed_copy(1)
        lneg = lnew.signed_copy(-1)
        # get part's signed lines
        slines = self.signlines
        lexists = False
        for sline in slines:
            if lpos == sline:
                lexists = True
                signline_new = sline
                break
            elif lneg == sline:
                lexists = True
                signline_new = sline.signed_copy(-1)
                signline_new.edge = False
                self.fea.register(signline_new)
                break
        else:
            # fired when we haven't broken out of the loop, the line is new
            if isinstance(lnew, geometry.SignLine):
                lnew = geometry.Line(lnew.pt(0), lnew.pt(1))
            self.fea.register(lnew)
            lnew.save_to_points()
            signline_new = lnew.signed_copy(1)
            self.fea.register(signline_new)
        signline_new.line.add_signline(signline_new)
        return [signline_new, lexists]

    def draw_circle(self, center_x, center_y, radius, num_arcs=4):
        """Draws a circle area and adds it to the part.

        Args:
            center_x (float): x-axis hole center
            center_y (float): y-axis hole center
            radius (float): hole radius
            num_arcs (int): number of arcs to use, must be >= 3

        Returns:
            loop (geometry.LineLoop): a LineLoop list of SignArc
        """
        center = geometry.Point(center_x, center_y)
        rvect = geometry.Point(0, radius)
        start = center + rvect
        self.goto(start.x, start.y)
        angles = np.linspace(360/num_arcs,360,num_arcs, endpoint=True)
        for ang in angles:
            point = geometry.Point(0, radius).rot_ccw_deg(ang)
            point = point + center
            self.draw_arc(point.x, point.y, center.x, center.y)
        loop = self.areas[-1].exlines
        self.__update()
        return loop

    def draw_hole(self, center_x, center_y, radius, num_arcs=4, filled=False):
        """Makes a hole in the part.

        Args:
            center_x (float): x-axis hole center
            center_y (float): y-axis hole center
            radius (float): hole radius
            num_arcs (int): number of arcs to use, must be >= 3
            filled (bool): whether to fill the hole

                * True: makes a new area in the part

        Returns:
            hole_lines (list or None): list of hole SignLine or SignArc

                * Returns None if hole was not made.
        """
        center = geometry.Point(center_x, center_y)
        area = self.__area_from_pt(center)
        if area == None:
            print("You can't make a hole here until there's an area here!")
            return None
        else:
            # make points
            rvect = geometry.Point(0, radius)
            start = center + rvect
            self.goto(start.x, start.y, holemode=True)
            angles = np.linspace(360/num_arcs,360,num_arcs, endpoint=True)
            for ang in angles:
                point = geometry.Point(0, radius).rot_ccw_deg(ang)
                point = point + center
                self.draw_arc(point.x, point.y, center.x, center.y)
            # make new area
            if filled:
                # reverse order, reverse sline directions, store in feamodel
                slines = list(area.holes[-1])
                slines.reverse()
                slines = [sline.signed_copy(-1) for sline in slines]
                slines = [self.__make_get_sline(sline)[0] for sline in slines]
                anew = self.fea.areas.append(geometry.Area(self, slines))
                self.areas.append(anew)
            self.__update()
            return area.holes[-1]

    def draw_arc_angle(self, degrees_ccw, center_x, center_y):
        """Makes an arc and adds it to the part.

        | Current point is the first arc point.
        | degrees_ccw is the swept angle in degrees, counterclockwise
        | (center_x, center_y) is the arc center

        | Degrees: Traversed angle of arc must be < 180 degrees

        Args:
            degrees_ccw (float): arc swept angle in degrees, counterclockwise
            center_x (float): arc center point x-coordinate
            center_y (float): arc center point y-coordinate

        Returns:
            list: [arc, arc_start_point, arc_end_point]
        """
        center = geometry.Point(center_x, center_y)
        radius_vector = self.__cursor - center
        radius_vector.rot_ccw_deg(degrees_ccw)
        end = center + radius_vector
        return self.draw_arc(end.x, end.y, center_x, center_y)

    def draw_arc(self, end_x, end_y, center_x, center_y):
        """Makes an arc and adds it to the part.

        | Current point is the first arc point.
        | (end_x, end_y) is the end point
        | (center_x, center_y) is the arc center

        | Degrees: Traversed angle of arc must be < 180 degrees
        | Radians: Traversed angle of arc must be < Pi

        Args:
            end_x (float): arc end point x-coordinate
            end_y (float): arc end point y-coordinate
            center_x (float): arc center point x-coordinate
            center_y (float): arc center point y-coordinate

        Returns:
            list: [arc, arc_start_point, arc_end_point]
        """
        pold = self.__cursor
        # make arc center point
        ctr = self.__make_get_pt(center_x, center_y)[0]
        # make arc end point
        self.__cursor = self.__make_get_pt(end_x, end_y)[0]

        # make arc
        arc = self.__make_get_sline(geometry.Arc(pold, self.__cursor, ctr))[0]

        if self.__holemode:
            area = self.__area_from_pt(self.__cursor)
            if area != None:
                closed = area.add_hole_sline(arc)
                if closed:
                    self.__holemode = False
            else:
                print('You must have a closed area here before making a hole!')
        else:
            self.areas[-1].add_sline(arc)
        return [arc, pold, self.__cursor]

    def draw_line_delta(self, delta_x, delta_y):
        """Draws a line a relative distance, and adds it to the part.

        Args:
            delta_x (float): x-axis delta distance to draw the line
            delta_y (float): y-axis delta distance to draw the line

        Returns:
            list: [line, point_start, point_end]
        """
        x = self.__cursor.x + delta_x
        y = self.__cursor.y + delta_y
        return self.draw_line_to(x, y)

    def draw_line_rad(self, dx_rad):
        """Draws a line a relative radial distance, and adds it to the part.

        Args:
            dx_rad (float): x-axis delta distance to draw the line

        Returns:
            list: [line, point_start, point_end]
        """
        return self.draw_line_delta(dx_rad, 0.0)

    def draw_line_ax(self, dy_ax):
        """Draws a line a relative axial distance, and adds it to the part.

        Args:
            dy_ax (float): y-axis delta distance to draw the line

        Returns:
            list: [line, point_start, point_end]
        """
        return self.draw_line_delta(0.0, dy_ax)

    def draw_line_to(self, x, y):
        """Draws a line to the given location, and adds it to the part.

        Args:
            x (float): x-axis coordinate of the end point
            y (float): y-axis coordinate of the end point

        Returns:
            list: [SignLine, point_start, point_end]
        """
        pold = self.__cursor
        self.__cursor = self.__make_get_pt(x, y)[0]
        sline = self.__make_get_sline(geometry.Line(pold, self.__cursor))[0]

        if self.__holemode:
            area = self.__area_from_pt(self.__cursor)
            if area != None:
                closed = area.add_hole_sline(sline)
                if closed:
                    self.__holemode = False
                    self.__update()
            else:
                print('You must have a closed area here before making a hole!')
        else:
            # drawing in last area
            self.areas[-1].add_sline(sline)
            # check for closure of the area
            if self.areas[-1].closed:
                self.__update()
        return [sline, pold, self.__cursor]

    def __get_maxlength(self):
        """Returns the max distance between points in the part."""
        points = self.points
        maxlen = 0.0
        # loop through points checking dist to next point
        for ind, point_1 in enumerate(points[:-1]):
            for point_2 in points[ind:]:
                vect = point_1 - point_2
                dist = vect.length()
                if dist > maxlen:
                    maxlen = dist
        return maxlen

    def __area_from_pt(self, point):
        """Returns the area that the point is inside.

        Args:
            point (Point): the point we are asking about

        Returns:
            Area or None:
                Area is the found area
                None is returned if the point is not in one of this part's areas
        """
        for area in self.areas:
            if area.contains_point(point):
                return area
        return None

    def fillet_lines(self, line1, line2, radius):
        """Fillets the given lines in the part.

        This inserts an arc in the part tangent to the two given lines.

        Args:
            line1 (SignLine): line that the arc starts on, arc is tangent
            line2 (SignLine): line that the arc ends on, arc is tangent
            radius (float): arc radius size

        Returns:
            list: [arc, start_point, end_point]
        """
        # check if the lines are touching
        if not line1.line.touches(line2.line):
            print('ERROR: Cannot fillet! Lines must touch!')
            return

        if line1.line.pt(1) == line2.line.pt(0):
            first_line = line1
            second_line = line2
        elif line2.line.pt(1) == line1.line.pt(0):
            first_line = line2
            second_line = line1
        else:
            print('ERROR: Sign lines must both be going in CW or CCW '
                  'direction. The two passed lines are going in '
                  'different directions. Unable to fillet them.')
            return

        tmp = self.__cursor
        # offset the lines, assuming area is being traced clockwise
        # get the intersection point
        magnitude = radius
        l1_off = first_line.offset(magnitude)
        l2_off = second_line.offset(magnitude)
        ctrpt = l1_off.intersects(l2_off)
        if ctrpt == None:
            # flip the offset direction if lines don't intersect
            magnitude = -radius
            l1_off = first_line.offset(magnitude)
            l2_off = second_line.offset(magnitude)
            ctrpt = l1_off.intersects(l2_off)

        # now we have an intersecting point
        p1_new = first_line.arc_tang_intersection(ctrpt, magnitude)
        p2_new = second_line.arc_tang_intersection(ctrpt, magnitude)
        rempt = first_line.pt(1)

        p1_new = self.__make_get_pt(p1_new.x, p1_new.y)[0]
        ctrpt = self.__make_get_pt(ctrpt.x, ctrpt.y)[0]
        p2_new = self.__make_get_pt(p2_new.x, p2_new.y)[0]

        # make the new arc
        arc = self.__make_get_sline(geometry.Arc(p1_new, p2_new, ctrpt))[0]

        # put the arc in the right location in the area
        area = first_line.lineloop.parent
        area.line_insert(first_line, arc)
        print('Arc inserted into area %i' % (area.id))

        # edit the adjacent lines to replace the removed pt
        first_line.set_pt(1, arc.pt(0))
        second_line.set_pt(0, arc.pt(1))
        # del old pt, store new points for the arc
        self.fea.points.remove(rempt)

        # reset the cursor to where it should be
        self.__cursor = tmp
        return [arc, arc.pt(0), arc.pt(1)]

    def fillet_all(self, radius):
        """Fillets all external lines not within 10 degrees of tangency

        Args:
            radius (float): the fillet radius to use

        Returns:
            arcs (list): list of SignArc
        """
        pairs = []
        for area in self.areas:
            for ind, sline in enumerate(area.exlines):
                prev_sline = area.exlines[ind-1]
                this_point = sline.pt(0)
                if len(this_point.lines) == 2:
                    # only fillet lines that are not shared by other areas
                    if (isinstance(sline, geometry.SignLine)
                        and isinstance(prev_sline, geometry.SignLine)):
                        # only fillet lines
                        perp1 = prev_sline.get_perp_vec(this_point)
                        perp2 = sline.get_perp_vec(this_point)
                        ang = perp1.ang_bet_deg(perp2)
                        is_tangent = (-10 <= ang <= 10)
                        if is_tangent == False:
                            pairs.append([prev_sline, sline])
        arcs = []
        for pair in pairs:
            arc = self.fillet_lines(pair[0], pair[1], radius)[0]
            arcs.append(arc)
        return arcs

    def label(self, axis):
        """Labels the part on a Matplotlib axis

        Args:
            axis (Matplotlib Axis): Matplotlib Axis
        """
        axis.text(self.center.y, self.center.x, self.get_name(),
                  ha='center', va='center')

    def plot(self, axis, label=True, color='yellow'):
        """Plots the part on the passed Matplotlib axis.

        Args:
            axis (Matplotlib axis): the axis we will plot on
            label (bool): True displays the part label
            color (tuple): r,g,b,a matplotlib color tuple
        """
        patches = []
        for area in self.areas:
            if area.closed:
                patches.append(area.get_patch())
        for patch in patches:
            patch.set_color(color)
            axis.add_patch(patch)
        # apply the label
        if label:
            self.label(axis)

    def __cut_line(self, point, line):
        """Cuts the passed line at the passed point.

        The passed line is cut into two lines. All areas that included the
        original line are updated.

        Args:
            line (Line or Arc): the line to cut, must be Line or Arc
            point (Point): the location on the line that we will cut it

        Returns:
            list: [pnew, lnew]
                pnew: the new point we created to cut the original line
                lnew: the new line we created, the end half of the orignal line
        """
        pnew = self.__make_get_pt(point.x, point.y)[0]
        if point.id != -1:
            # if passed point already exists, use it
            pnew = point
        pend = line.pt(1)
        line.set_pt(1, pnew) # shortens the line
        new_prim = geometry.Line(pnew, pend)
        if isinstance(line, geometry.Arc):
            new_prim = geometry.Arc(pnew, pend, line.actr)
        new_sline = self.__make_get_sline(new_prim)[0]
        # insert the new line into existing areas
        is_line = isinstance(line, geometry.Line) or isinstance(line, geometry.Arc)
        is_sline = isinstance(line, geometry.SignLine) or isinstance(line, geometry.SignArc)
        print('Cutting line (is_line, is_sline, signlines) (%s, %s, %i)' % (is_line, is_sline, len(line.signlines)))
        slines = line.signlines
        for sline in slines:
            area = sline.lineloop.parent
            if sline.sign == 1:
                # cutting line in clockwise area, where line is pos
                area.line_insert(sline, new_sline)
            elif sline.sign == -1:
                # cutting line in clockwise area, where line is neg
                rev_sline = self.__make_get_sline(new_sline.signed_copy(-1))[0]
                area.line_insert(sline, rev_sline, after=False)
        return [pnew, new_sline]

    def __cut_area(self, area, start_pt, end_pt):
        """Cuts the part area from start_pt to end_pt."""
        # select the line portions that define the areas
        # list[:low] excludes low index
        # list [high:] includes high index
        #    we want the line which start with the point
        lpre_start = area.line_from_startpt(start_pt)
        lpre_end = area.line_from_startpt(end_pt)
        if lpre_start == None or lpre_end == None:
            self.fea.plot_geometry()
            print(area.exlines)
        istart = area.exlines.index(lpre_start)
        iend = area.exlines.index(lpre_end)
        low = min(istart, iend)
        high = max(istart, iend)

        # lists of lines for areas
        beg = area.exlines[:low]
        mid = area.exlines[low:high]
        end = area.exlines[high:]

        # make cut line for [beg + cut + end] area
        start_pt = mid[0].pt(0)
        end_pt = mid[-1].pt(1)
        fwd = geometry.Line(start_pt, end_pt)
        rev = geometry.Line(end_pt, start_pt)

        # update existing area
        cline = self.__make_get_sline(fwd)[0]
        alist_curr = beg + [cline] + end
        area.update(alist_curr)

        # make new area
        cline_rev = self.__make_get_sline(rev)[0]
        alist_other = mid + [cline_rev]
        anew = geometry.Area(self, alist_other)
        self.fea.register(anew)
        self.areas.append(anew)

        # fix holes
        self.__store_holes()

    def __merge_hole(self, area, start_pt, end_pt):
        """Merges the hole into its area with a line between passed points."""
        # line will be drawn from start point on exlines to end point on hole
        hole_points = area.holepoints
        if start_pt in hole_points:
            tmp = start_pt
            start_pt = end_pt
            end_pt = tmp
        lpre_start = area.line_from_startpt(start_pt)
        hole_line = area.line_from_startpt(end_pt)
        if lpre_start == None or hole_line == None:
            self.fea.plot_geometry()
        ind = area.exlines.index(lpre_start)

        # store sections of the area
        beg = area.exlines[:ind]
        end = area.exlines[ind:]
        thehole = None
        mid = []
        for hole in area.holes:
            for sline in hole:
                if sline == hole_line:
                    ind = hole.index(sline)
                    mid = hole[ind:] + hole[:ind]
                    thehole = hole
                    break
            if mid != []:
                break

        fwd = geometry.Line(start_pt, end_pt)
        fwd_sline = self.__make_get_sline(fwd)[0]
        rev_sline = fwd_sline.signed_copy(-1)
        self.fea.register(rev_sline)
        rev_sline.line.add_signline(rev_sline)
        alist_curr = beg + [fwd_sline] + mid + [rev_sline] + end
        area.holes.remove(thehole)
        area.update(alist_curr)

    def __get_cut_line(self, cutline):
        """Returns a cut line beginning and ending on the part."""
        # find all intersections
        lines = self.lines
        points = set()

        # add line intersections
        for line in lines:
            newpt = line.intersects(cutline)
            if newpt != None:
                points.add(newpt)

        # loop through intersection points, storing distance
        points = list(points)
        for (ind, point) in enumerate(points):
            dist = point - cutline.pt(0)
            dist = dist.length()
            pdict = {'dist': dist, 'point': point}
            points[ind] = pdict

        # sort the points by dist, lowest to highest, return first cut
        points = sorted(points, key=lambda k: k['dist'])
        start = points[0]['point']
        end = points[1]['point']
        new_cut = geometry.Line(start, end)
        return new_cut

    def __cut_with_line(self, cutline, debug):
        """Cuts the part using the passed line.

        Args:
            cutline (Line): line to cut the area with
            debug (list): bool for printing, bool for plotting after every cut
        """
        # find all intersections
        lines = self.lines
        points = set()

        # add line intersections
        for line in lines:
            if debug[0]:
                print('Checking X between %s and cutline' % line.get_name())
            newpt = line.intersects(cutline)
            if debug[0]:
                print(' Intersection: %s' % newpt)
            if newpt != None:
                points.add(newpt)

        # loop through intersection points, storing distance and lines to cut
        points = list(points)
        for (ind, point) in enumerate(points):
            dist = point - cutline.pt(0)
            dist = dist.length()
            pdict = {'dist': dist}
            realpt = self.__get_point(point)
            # we only want to store lines to cut here
            if realpt == None or realpt.arc_center == True:
                # we could have an existing arc center on a line that needs to
                # be cut
                if realpt == None:
                    realpt = point
                for line in lines:
                    point_on_line = line.coincident(realpt)
                    if point_on_line and point not in line.points:
                        pdict['line'] = line
                        break
            pdict['point'] = realpt
            points[ind] = pdict

        # sort the points by dist, lowest to highest
        points = sorted(points, key=lambda k: k['dist'])
        if debug[0]:
            print('==================================')
            print('Points on the cutline!------------')
            for pdict in points:
                print(pdict['point'])
                print(' dist %.3f' % pdict['dist'])
                if 'line' in pdict:
                    print(' X cut line: %s' % pdict['line'])
            print('==================================')

        # loop through the points cutting areas
        for ind in range(len(points)):
            pdict = points[ind]
            start_pt = pdict['point']
            if 'line' in pdict:
                # cut the line and point to the real new point
                print('Cut through line %s' % pdict['line'].get_name())
                pnew = self.__cut_line(start_pt, pdict['line'])[0]
                points[ind]['point'] = pnew
                start_pt = pnew
            end_pt = None
            pavg = None
            area = None
            if ind > 0:
                # find the area we're working on
                end_pt = points[ind-1]['point']
                pavg = start_pt + end_pt
                pavg = pavg*0.5
                area = self.__area_from_pt(pavg)

                if area == None:
                    # stop cutting if we are trying to cut through a holes
                    print('No area found at point avg, no cut made')
                    break
                start_hole = start_pt in area.holepoints
                end_hole = end_pt in area.holepoints
                if start_hole and end_hole and area != None:
                    print('Trying to join holes, no cut made')
                    break
                    # stop cutting if we are trying to join holes
                if end_hole == True or start_hole == True:
                    print('Merging hole in %s' % area.get_name())
                    self.__merge_hole(area, start_pt, end_pt)
                else:
                    print('Cutting %s' % area.get_name())
                    self.__cut_area(area, start_pt, end_pt)
                if debug[1]:
                    self.fea.plot_geometry()

    def __store_holes(self):
        """Puts all holes in their correct areas"""
        holes = []
        for area in self.areas:
            holes += area.holes
        for hole in holes:
            hole_area = hole.parent
            for area in self.areas:
                is_inside = hole.inside(area.exlines)
                if is_inside == True:
                    if area != hole_area:
                        # delete the hole from the old area, move it to the new
                        hole.set_parent(area)
                        hole_area.holes.remove(hole)
                        hole_area.close()
                        area.holes.append(hole)
                        area.close()
                        afrom, ato = hole_area.get_name(), area.get_name()
                        print('Hole moved from %s to %s' % (afrom, ato))

    def __vect_to_line(self, point, cvect):
        """Returns a cutting line at a given point and cutting vector.

        Args:
            point (Point): the location we are cutting from
            cvect (Point): the vector direction of the cut from pt

        Returns:
        cutline (Line): cut line
        """
        cvect.make_unit()
        vsize = self.__get_maxlength()
        endpt = point + cvect*vsize
        cutline = geometry.Line(point, endpt)
        cutline = self.__get_cut_line(cutline)
        return cutline

    def __chunk_area(self, area, mode, exclude_convex, debug):
        """Cuts the passed area into regular smaller areas.

        The cgx mesher only accepts areas which are 3-5 sides
        so one may need to call this before using that mesher.
        Cuts are made perpendicular to tangent points or at
        internal corners.
        At internal corners two perpendicular cuts are made.

        Args:
            area (Area): the area to cut into smaller areas
            mode (str): 'both', 'holes' or 'ext' chunks the area using the
                points form this set. See part.chunk
            exclude_convex (bool): If true exclude cutting convex tangent points
            debug (list): bool for printing, bool for plotting after every cut
        """
        # store the cuts first, then cut after
        cuts = [] # each item is a dict with a pt and vect in it
        loops = []
        cut_point_sets = []
        if mode == 'holes':
            loops = area.holes
        elif mode == 'ext':
            loops = [area.exlines]
        elif mode == 'both':
            loops = area.holes + [area.exlines]
        for loop in loops:
            for ind, line in enumerate(loop):
                line_pre = loop[ind-1]
                line_post = line
                point = line_pre.pt(1)
                perp1 = line_pre.get_perp_vec(point)
                perp2 = line_post.get_perp_vec(point)
                #tan1 = line_pre.get_tan_vec(point)
                #tan2 = line_post.get_tan_vec(point)
                # flip these vectors later to make them cut the area(s)
                ang = perp1.ang_bet_deg(perp2)
                cut = {}
                make_cut = True
                pre_arc = isinstance(line_pre, geometry.SignArc)
                post_arc = isinstance(line_post, geometry.SignArc)
                if pre_arc or post_arc:
                    if pre_arc and post_arc:
                        if (line_pre.concavity == 'convex'
                            and line_post.concavity == 'convex'
                            and exclude_convex == True):
                                make_cut = False
                    else:
                        # only one is an arc
                        arc = line_pre
                        if post_arc:
                            arc = line_post
                        if arc.concavity == 'convex' and exclude_convex == True:
                            make_cut = False
                is_tangent = (-10 <= ang <= 10)
                is_int_corner = (45 <= ang <= 135)
                """
                print('-------------------')
                print('%s' % point)
                print('Angle is %.3f' % ang)
                print('Make cut %s' % make_cut)
                print('is_tangent %s' % is_tangent)
                print('is_int_corner %s' % is_int_corner)
                """
                if is_tangent:
                    if make_cut == True:
                        # tangent
                        cut = {'pt':point, 'vect':perp1*-1}
                        cut_line = self.__vect_to_line(cut['pt'], cut['vect'])
                        pset = set(cut_line.points)
                        if pset not in cut_point_sets:
                            cut_point_sets.append(pset)
                            cut['line'] = cut_line
                            cuts.append(cut)
                elif is_int_corner:
                    # internal corner
                    cut = {'pt':point, 'vect':perp1*-1}
                    cut_line = self.__vect_to_line(cut['pt'], cut['vect'])
                    pset = set(cut_line.points)
                    if pset not in cut_point_sets:
                        cut_point_sets.append(pset)
                        cut['line'] = cut_line
                        cuts.append(cut)
                    cut = {'pt':point, 'vect':perp2*-1}
                    cut_line = self.__vect_to_line(cut['pt'], cut['vect'])
                    pset = set(cut_line.points)
                    if pset not in cut_point_sets:
                        cut_point_sets.append(pset)
                        cut['line'] = cut_line
                        cuts.append(cut)
                elif ang < 0:
                    # external corner
                    # do not split these
                    pass

        # do the cuts
        for cut in cuts:
            print('--------------------')
            print('Cut point:', cut['pt'].get_name())
            print('Cut line:', cut['line'])
            self.__cut_with_line(cut['line'], debug)

    def chunk(self, mode='both', exclude_convex = True, debug=[0, 0]):
        """Chunks all areas in the part.

        Args:
            mode (str): area chunking mode

                - 'both': cuts areas using holes and exterior points
                - 'holes': cut areas using holes points only
                - 'ext': cut areas using exterior points only

            exclude_convex (bool): If true exclude cutting convex tangent points
        """
        for area in self.areas:
            if area.closed:
                min_sides = 5
                has_holes = len(area.holes) > 0
                ext_gr = len(area.exlines) > min_sides
                both_false = (has_holes == False and ext_gr == False)
                if mode == 'holes' and has_holes:
                    self.__chunk_area(area, mode, exclude_convex, debug)
                elif (mode == 'both'
                      and (has_holes or ext_gr or not exclude_convex)):
                    self.__chunk_area(area, mode, exclude_convex, debug)
                elif mode == 'ext' and (ext_gr or not exclude_convex):
                    self.__chunk_area(area, mode, exclude_convex, debug)
                else:
                    aname = area.get_name()
                    val = 'Area %s was not chunked because it had' % aname
                    adder = ''
                    if mode == 'both' and both_false:
                        adder = '<= %i lines and no holes.' % min_sides
                    elif has_holes == False and (mode in ['both', 'holes']):
                        adder = 'no holes.'
                    elif ext_gr == False and (mode in ['both', 'ext']):
                        adder = '<= %i lines.' % min_sides
                    print('%s %s' % (val, adder))

        # store the left, right, top, and bottom lines
        self.__update()

    def __update(self):
        """Updates the left, right, top, bottom sides and area and center"""
        self.__set_side('left')
        self.__set_side('right')
        self.__set_side('top')
        self.__set_side('bottom')
        self.area, self.center = self.__calc_area_center()

    def __str__(self):
        """Returns string listing object type, id number and name."""
        val = 'Part, id=%i name=%s' % (self.id, self.get_name())
        return val



"""This module stores the Selector object

Selector is used in FeaModel to store the model's selected set.
"""

from . import partmodule
from . import geometry
from . import mesh
from . import components

class Selector(object):
    """Makes a selector which stores the selected items in the FeaModel.

    Args:
        feamodel (FeaModel): the parent FeaModel object

    Attributes:
        __fea (FeaModel): the parent FeaModel
        __all (bool): If everything is selected all=True
        parts (set): currently selected parts
        areas (set): currently selected areas
        lines (set): currently selected signed lines or arcs
        points (set): currently selected points
        elements (set): currently selected elements
        faces (set): currently selected faces
        nodes (set): currently selected nodes
        __parts (set): internal storage set of parts
        __areas (set): internal storage set of areas
        __slines (set): internal storage set of signed lines and arcs
        __points (set): internal storage set of points
        __elements (set): internal storage set of elements
        __faces (set): internal storage set of faces
        __nodes (set): internal storage set of nodes
    """
    def __init__(self, feamodel):
        self.__fea = feamodel
        self.__all = True
        self.__parts = set()
        self.__areas = set()
        self.__slines = set()
        self.__points = set()
        self.__elements = set()
        self.__faces = set()
        self.__nodes = set()

    @property
    def parts(self):
        """Returns selected parts."""
        if self.__all:
            return self.__fea.parts
        else:
            return self.__parts

    @property
    def areas(self):
        """Returns selected areas."""
        if self.__all:
            return self.__fea.areas
        else:
            return self.__areas

    @property
    def lines(self):
        """Returns selected signed lines or arcs."""
        if self.__all:
            return self.__fea.signlines
        else:
            return self.__slines

    @property
    def points(self):
        """Returns selected points."""
        if self.__all:
            points = [pnt for pnt in self.__fea.points if pnt.arc_center == False]
            return points
        else:
            return self.__points

    @property
    def elements(self):
        """Returns selected elements."""
        if self.__all:
            return self.__fea.elements
        else:
            return self.__elements

    @property
    def faces(self):
        """Returns selected faces."""
        if self.__all:
            return self.__fea.faces
        else:
            return self.__faces

    @property
    def nodes(self):
        """Returns selected nodes."""
        if self.__all:
            return self.__fea.nodes
        else:
            return self.__nodes

    @staticmethod
    def __add_items(inclusive, items, item, parent_list, this_set):
        """Adds an item to items if appropriate.

        Helper function for select_above_all
        """
        if inclusive == False:
            items.add(item)
        else:
            parent_set = set(parent_list)
            if parent_set.issubset(this_set):
                items.add(item)

    def __add_select(self, items):
        """Adds items to the selection set."""
        for item in items:
            if isinstance(item, partmodule.Part):
                self.__parts.add(item)
            elif isinstance(item, geometry.Area):
                self.__areas.add(item)
            elif (isinstance(item, geometry.SignLine)
                  or isinstance(item, geometry.SignArc)):
                self.__slines.add(item)
            elif isinstance(item, geometry.Point):
                self.__points.add(item)
            elif isinstance(item, mesh.Element):
                self.__elements.add(item)
            elif isinstance(item, mesh.Face):
                self.__faces.add(item)
            elif isinstance(item, mesh.Node):
                self.__nodes.add(item)

    def allsel_under(self, sel_type, full=True, byfaces=False):
        """Selects all progeny under current sel_type items.

        Args:
            sel_type (str): the item type to select under

                - 'parts'
                - 'areas'
                - 'lines'
                - 'points' (needs full=True)
                - 'elements'
                - 'faces'

            full (bool): if False selects under one branch

                - solid modeling branch: area->line->keypoints->points
                - mesh branch: elements->faces->nodes

                Cross branch connections:
                - faces are under lines
                - elements are under areas
                - nodes are under lines and points

                If True, selects all items connected to sub items:
                - Parts:

                    - Areas
                    - Elements, element faces + element nodes
                    - Lines, line points

                - Areas:

                    - Elements, element faces + element nodes
                    - Lines, line points

                -Lines:

                    - Faces, nodes
                    - byfaces=True: elements above faces
                    - byfaces=False: elements above line nodes selected

                -Points:

                    - Nodes
                    - byfaces=True: no elements selected
                    - byfaces=False: elements above point nodes selected

            byfaces (bool): If true, the child elements will only be selected
                if a parent face was selected. If false just an element node
                needs to be selected to select the element.

                -This is only applicable when sel_type == 'lines'
        """
        sets = ['parts', 'areas', 'lines', 'points', 'elements', 'faces']
        if isinstance(sel_type, str):
            if sel_type in sets:
                this_set = getattr(self, sel_type)
                for item in this_set:
                    if isinstance(item, partmodule.Part):
                        # areas
                        self.select_below_all('parts')
                        # slines
                        self.select_below_all('areas')
                        # points
                        self.select_below_all('lines')
                        # select related elements if full is on
                        if full:
                            elements = item.elements
                            self.select(elements, also=True)
                            # sel faces
                            self.select_below_all('elements')
                            # sel nodes
                            self.select_below_all('faces')

                    elif isinstance(item, geometry.Area):
                        # slines
                        self.select_below_all('areas')
                        # points
                        self.select_below_all('lines')
                        # select related elements if full is on
                        if full:
                            elements = item.elements
                            self.select(elements, also=True)
                            # sel faces
                            self.select_below_all('elements')
                            # sel nodes
                            self.select_below_all('faces')
                    elif (isinstance(item, geometry.SignLine)
                          or isinstance(item, geometry.SignArc)):
                        # sel points
                        self.select_below_all('lines')
                        # select related faces, nodes, and elements
                        if full:
                            # select faces
                            faces = item.faces
                            self.select(faces, also=True)
                            # sel nodes
                            self.select_below_all('faces')
                            # byfaces logic
                            if byfaces:
                                # sel elements above faces
                                self.select_above_all('faces', inclusive=False)
                            else:
                                # select elements connected to nodes
                                elements = set()
                                for node in self.nodes:
                                    elements.update(node.elements)
                                self.select(elements, also=True)
                    elif isinstance(item, geometry.Point):
                        if full:
                            # sel nodes and elements
                            self.select(item.nodes, also=True)
                            if byfaces == False:
                                for node in item.nodes:
                                    self.select(node.elements, also=True)
            else:
                print('The passed sel_type must be a valid type.')
        else:
            print('The passed sel_type must be a string!')

    def select_below(self):
        """Selects children below currently selected items.

        Only one set must be selected.
        """
        sets = ['parts', 'areas', 'lines', 'points', 'elements', 'faces',
                'nodes']
        sel_sets = []
        for set_name in sets:
            this_set = getattr(self, set_name)
            if len(this_set) > 0:
                sel_sets.append(set_name)
        # check if number of visible sets is one
        if len(sel_sets) == 0 or len(sel_sets) > 1:
            print('Only one set must be selected to use this function.')
        elif len(sel_sets) == 1:
            sel_type = sel_sets[0]
            if sel_type == 'nodes':
                print("One can't select anything below nodes!'")
            else:
                self.select_below_all(sel_type)

    def select_below_all(self, sel_type):
        """Selects children below all currently selected items of sel_type.

        Args:
            sel_type (str): type to select items below

                - 'parts': selects areas
                - 'areas': selects lines (type SignLine and SignArc)
                - 'lines': selects points
                - 'elements': selects faces
                - 'faces': selects nodes
        """
        items = set()
        if sel_type == 'parts':
            for part in self.parts:
                for area in part.areas:
                    items.add(area)
        elif sel_type == 'areas':
            for area in self.areas:
                items.update(area.signlines)
        elif sel_type == 'lines':
            for sline in self.lines:
                items.update(sline.line.points)
        elif sel_type == 'elements':
            for element in self.elements:
                items.update(element.faces)
        elif sel_type == 'faces':
            for face in self.faces:
                items.update(face.nodes)
        self.__add_select(items)

    def select_above(self, inclusive=False):
        """Selects parents above currently selected items.

        Only one set must be selected.
        """
        sets = ['parts', 'areas', 'lines', 'points', 'elements', 'faces',
                'nodes']
        sel_sets = []
        for set_name in sets:
            this_set = getattr(self, set_name)
            if len(this_set) > 0:
                sel_sets.append(set_name)
        # check if number of visible sets is one
        if len(sel_sets) == 0 or len(sel_sets) > 1:
            print('Only one set must be selected to use this function.')
        elif len(sel_sets) == 1:
            sel_type = sel_sets[0]
            if sel_type == 'parts':
                print("One can't select anything above parts!'")
            else:
                self.select_above_all(sel_type, inclusive)

    def select_above_all(self, sel_type, inclusive=False):
        """Selects parents above all currently selected items of sel_type.

        Args:
            sel_type (str): type to select items above

                - 'areas': selects parts
                - 'lines': selects areas
                - 'points': selects signlines or signarcs
                - 'faces': selects elements
                - 'nodes': selects faces

            inclusive (bool): if True, all child entities must be selected for
                the parent to be selected. If false all parents connected to
                the child are selected. Exampes:

                - 'areas' + True: all part areas must be selected to select part
                - 'areas' + False: parts with selected areas in them are selected
                - 'lines' + True: all lines must be selected to select area
                - 'lines' + False: areas with selected lines in them are selected
        """
        items = set()
        if sel_type == 'areas':
            this_set = self.areas
            for area in this_set:
                item = area.parent
                parent_list = area.parent.areas
                self.__add_items(inclusive, items, item, parent_list, this_set)
        elif sel_type == 'lines':
            this_set = self.lines
            for sline in this_set:
                item = sline.parent
                parent_list = sline.parent.signlines
                self.__add_items(inclusive, items, item, parent_list, this_set)
        elif sel_type == 'points':
            this_set = set(self.points)
            for point in this_set:
                for parent_line in point.lines:
                    for sline in parent_line.signlines:
                        item = sline
                        parent_list = sline.points
                        self.__add_items(inclusive, items, item, parent_list,
                                         this_set)
        elif sel_type == 'faces':
            this_set = set(self.faces)
            for face in this_set:
                item = face.element
                parent_list = item.faces
                self.__add_items(inclusive, items, item, parent_list, this_set)
        elif sel_type == 'nodes':
            this_set = set(self.nodes)
            for node in this_set:
                for face in node.faces:
                    item = face
                    parent_list = face.nodes
                    self.__add_items(inclusive, items, item, parent_list,
                                     this_set)
        self.__add_select(items)

    def allselect(self):
        """Selects all items."""
        self.select(items='all')

    def select(self, items='', also=False):
        """Selects an item or a list of items.

        Agrs:
            items (str or item or list): item(s) to select
            also (bool): whether to add the items to the current selection set

                - True: add the items to the current selection set
                - False: only select the passed items
        """
        # empty the selection set if also == False
        if items == '':
            print('You must pass in an item or item(s)!')
        else:
            if also == False:
                self.select_none()
            # Turn on 'all' mode if items == None, otherwise, get the items
            if items == 'all':
                self.__all = True
            else:
                items = self.__fea.get_items(items)
                for (ind, item) in enumerate(items):
                    # if a component was passed, get its child items instead
                    if isinstance(item, components.Component):
                        real_items = item.get_children()
                        items[ind] = real_items
                self.__add_select(items)

    def select_all(self, sel_type='all', also=False):
        """Selects all items of sel_type from the full feamodel.

        All items currently selected items are removed from the selection set
        before the all set is selected.

        Args:
            sel_type (str): the items to select

                - 'all'
                - 'parts'
                - 'areas'
                - 'lines'
                - 'points'
                - 'elements'
                - 'faces'
                - 'nodes'
            also (bool): if False, empties the selection set before selecting
                all items, if True adds new items to the current set
        """
        if also == False:
            self.__all = False
            self.select_none()
        # now select all items of type
        if sel_type == 'all':
            self.select()
        elif sel_type == 'parts':
            self.__add_select(self.__fea.parts)
        elif sel_type == 'areas':
            self.__add_select(self.__fea.areas)
        elif sel_type == 'lines':
            self.__add_select(self.__fea.signlines)
        elif sel_type == 'points':
            points = [pnt for pnt in self.__fea.points if pnt.arc_center == False]
            self.__add_select(points)
        elif sel_type == 'elements':
            self.__add_select(self.__fea.elements)
        elif sel_type == 'faces':
            self.__add_select(self.__fea.faces)
        elif sel_type == 'nodes':
            self.__add_select(self.__fea.nodes)

    def deselect(self, items):
        """Removes the passed item or items from the selection set.
        """
        self.__all = False
        for item in items:
            if isinstance(item, partmodule.Part):
                self.__parts.discard(item)
            elif isinstance(item, geometry.Area):
                self.__areas.discard(item)
            elif (isinstance(item, geometry.SignLine)
                  or isinstance(item, geometry.SignArc)):
                self.__slines.discard(item)
            elif isinstance(item, geometry.Point):
                self.__points.discard(item)
            elif isinstance(item, mesh.Element):
                self.__elements.discard(item)
            elif isinstance(item, mesh.Face):
                self.__faces.discard(item)
            elif isinstance(item, mesh.Node):
                self.__nodes.discard(item)

    def deselect_all(self, sel_type):
        """Deselects all items of sel_type.

        All items currently selected items are removed from the selection set
        before the all set is selected.

        Args:
            sel_type (str): the items to select

                - 'all'
                - 'parts'
                - 'areas'
                - 'lines'
                - 'points'
                - 'elements'
                - 'faces'
                - 'nodes'
        """
        if self.__all == True:
            self.__parts = self.parts
            self.__areas = self.areas
            self.__slines = self.lines
            self.__points = self.points
            self.__elements = self.elements
            self.__faces = self.faces
            self.__nodes = self.nodes
            self.__all = False
        if sel_type == 'all':
            self.select_none()
        elif sel_type == 'parts':
            self.__parts = set()
        elif sel_type == 'areas':
            self.__areas = set()
        elif sel_type == 'lines':
            self.__slines = set()
        elif sel_type == 'points':
            self.__points = set()
        elif sel_type == 'elements':
            self.__elements = set()
        elif sel_type == 'faces':
            self.__faces = set()
        elif sel_type == 'nodes':
            self.__nodes = set()

    def select_none(self):
        """Empties out the selection object.
        """
        self.__all = False
        self.__parts = set()
        self.__areas = set()
        self.__slines = set()
        self.__points = set()
        self.__elements = set()
        self.__faces = set()
        self.__nodes = set()

    def print_summary(self):
        """Prints a summary of the number so items selected in each sel_type.
        """
        sets = ['parts', 'areas', 'lines', 'points', 'elements', 'faces',
                'nodes']
        spacer = '----------------------------------'
        print(spacer)
        print("View's currently selected items:")
        for sel_type in sets:
            items = getattr(self, sel_type)
            print(' %s: %i selected' % (sel_type, len(items)))
        print(spacer)




""" This module stores the FeaModel class.

This class is the highest level object in a pycalculix program.
It stores all parts, loads, constraints, mesh, problem, and results_file
objects.
"""

import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection  # element plotting
import matplotlib.colors as colors
import matplotlib.cm as cmx
import subprocess # used to launch meshers cgx and gmsh
import os # used to delete files written by cgx
from numpy.core.function_base import linspace # need to make contours

from . import environment
from . import base_classes
from . import geometry
from . import material
from . import components
from . import connectors
from . import loads
from . import mesh
from . import partmodule
from . import problem
from . import selector

# element colors, 0-1, 0=black, 1=whate
ECOLOR = '.4'
FCOLOR = '.9'
CMAP = 'jet' # for results
GEOM_CMAP = 'Pastel1' # color map for parts or areas

class FeaModel(object):
    """Makes a FeaModel instance.

    Parts, area, lines, arcs, points, nodes, elements, faces, models,
    components, and loads are stored in this object.

    Args:
        model_name (str): project name for this FeaModel, this is a file prefix
        ccx (None or str): path to Calculix ccx solver, pass this when you want
            to overwite the default program location.
            None means the default envionment.CCX is used.
        cgx (None or str): path to Calculix cgx mesher, pass this when you want
            to overwite the default program location.
            None means the default envionment.CGX is used.
        gmsh (None or str): path to gmsh mesher, pass this when you want
            to overwite the default program location.
            None means the default envionment.GMSH is used.

    Attributes:
        fname (str): FeaModel project file name prefix
        points (Itemlist): list of all Point
        lines (Itemlist): list of all Line and Arc
        signlines (Itemlist): list of all SignLine and SignArc
        lineloops (Itemlist): list of all LineLoop, each contains SignLine SignArc
        areas (Itemlist):list of all Area, each contains LineLoop(s)
        parts (Itemlist): list of all Part
        matls (Itemlist): list of all materials
        components (Itemlist): list of all components
        loads (dict): a dictionary of loads

            | Key (float): the load time point
            | loads[time] = list of loads for that time step
            | See method set_time to change the current time.
            | Time = 0.0 stores constant loads, such as:
            |    material, thickness

        contacts (Itemlist): list of contacts
        surfints (Itemlist): list of surface interactions
        problems (Itemlist): list of problems
        nodes (Meshlist): list of all mesh nodes
        eshape (str): element shape

            - 'quad': quadrilateral elements (Default)
            - 'tri': triangle elements
        eorder (int): element order, 1 or 2

            - 1: elements meshed with corner nodes only
            - 2: (Default) elements meshed with corner and midside nodes
        elements (Meshlist): list of all mesh elements
        faces (list): list of all element faces, includes non-exterior ones
        view (Selector): currently selected items

            Attributes:
                parts: list of selected parts
                areas: list of selected areas
                lines: list of selected signed lines or arcs
                points: list of selected points
                elements: list of selected elements
                faces: list of selected faces
                nodes: list of selected nodes
        time (float): current model time value, defaults to 1.0
        units (dict): a dict to store units for different fields

            Keys:
                - 'displ': displacement or location
                - 'force': force
                - 'stress': stress
                - 'temp': temperature
                - 'density': density (mass/volume)
                - 'time': time

            Returns:
                Text string describing the units for the given key field.

                See the set_units method.

                For example when units have been set to metric, the below
                values are returned.

                - 'dist' --> 'm'
                - 'density' --> 'kg/(m^3)'
                - 'time' --> 's'
    """
    def __init__(self, model_name, ccx=None, cgx=None, gmsh=None):
        self.fname = model_name
        self.points = base_classes.Itemlist()
        self.lines = base_classes.Itemlist()
        self.signlines = base_classes.Itemlist()
        self.lineloops = base_classes.Itemlist()
        self.areas = base_classes.Itemlist()
        self.parts = base_classes.Itemlist()
        self.matls = base_classes.Itemlist()
        self.components = base_classes.Itemlist()
        self.loads = {}    # store loads by time
        self.contacts = base_classes.Itemlist()
        self.surfints = base_classes.Itemlist()
        self.problems = base_classes.Itemlist()
        self.nodes = base_classes.Meshlist()
        self.elements = base_classes.Meshlist()
        self.faces = []
        self.view = selector.Selector(self)
        self.time = 1.0 # 0 is model set-up, 1 is first step, etc
        self.eshape = 'quad'
        self.eorder = 2
        self.units = {}

        # fix the paths to the needed programs, ccx, cgx, and gmsh
        if ccx != None:
            environment.CCX = ccx
        if cgx != None:
            environment.CGX = cgx
        if gmsh != None:
            environment.GMSH = gmsh

    def set_ediv(self, items, ediv):
        """Sets the number of elements on the passed line.

        Args:
            items (str or SignLine or SignArc or list): lines to set ediv on

                - str: 'L0'
                - list of str ['L0', 'L1']
                - list of SignLine or SignArc part.bottom or part.hole[-1]

            ediv (int): number of elements to mesh on the line
        """
        items = self.get_items(items)
        for line in items:
            line.set_ediv(ediv)

    def set_esize(self, items, esize):
        """Sets the element size on the passed line.

        Args:
            items (str or SignLine or SignArc or list): lines or points to set esize on

                - str: 'L0'
                - list of str ['L0', 'L1', 'P3']
                - list of SignLine or SignArc part.bottom or part.hole[-1]

            esize (float): size of the mesh elements on the line
        """
        items = self.get_items(items)
        for item in items:
            item.set_esize(esize)

    def set_units(self, dist_unit='m', cfswitch=False):
        """Sets the units that will be displayed when plotting.

        Picks a unit set based on the passed distance unit.
        That unit set is printed to the console when set.
        Defaults to MKS units (meter-kilogram-second)

        ========  =====  ======  ===========  ===============  ====
        Distance  Force  Stress  Temperature  Density          Time
        ========  =====  ======  ===========  ===============  ====
        'm'       'N'    'Pa'    'K'          'kg/(m^3)'       's'
        'mm'      'N'    'MPa'   'K'          'tonne/(mm^3)'   's'
        'in'      'lbf'  'psi'   'R'          'slinch/(in^3)'  's'
        'ft'      'lbf'  'psf'   'R'          'slug/(ft^3)'    's'
        ========  =====  ======  ===========  ===============  ====

        See get_units method or returning text strings based on unit types.

        Args:
            dist_unit (str): string of distance unit. Options:

                - 'm': meter
                - 'mm': milimeter
                - 'in': inch
                - 'ft': foot

            cfswitch (bool): Celsius/Fahrenheit temperature switch.
                Default is False.

                If True, this switches temperature from K-->C or from R-->F
                Default keeps units as shown in above table.
        """
        keys = ['dist', 'force', 'stress', 'temp', 'density', 'time']

        m_newton = ['m', 'N', 'Pa', 'K', 'kg/(m^3)', 's']
        mm_newton = ['mm', 'N', 'MPa', 'K', 'tonne/(mm^3)', 's']
        in_lbf = ['in', 'lbf', 'psi', 'R', 'slinch/(in^3)', 's']
        ft_lbf = ['ft', 'lbf', 'psf', 'R', 'slug/(ft^3)', 's']
        unit_systems = [m_newton, mm_newton, in_lbf, ft_lbf]
        vals = [usys for usys in unit_systems if usys[0] == dist_unit][0]

        if cfswitch:
            # switches K-->C and R-->F
            newunit = {'K':'C', 'R':'F'}
            vals['temp'] = newunit[vals['temp']]

        # set values
        adict = dict(zip(keys, vals))
        adict['displ'] = adict['dist']
        print('================================================')
        print('Units have been set to %s_%s' % (adict['dist'], adict['force']))
        for key in adict:
            print('For %s use %s' % (key, adict[key]))
        print('================================================')
        self.units = adict

    def get_units(self, *args):
        """Returns units for the passed arguments. Accepts and returns a list
        of strings.

        Options for inputs:
            - 'dist'
            - 'displ' (same units as 'dist')
            - 'force'
            - 'stress'
            - 'temp'
            - 'density'
            - 'time'
        """
        res = []
        for arg in args:
            mystr = ''
            if arg in self.units:
                # check if the requested item is in the units dict
                mystr = ' ('+ self.units[arg] + ')'
            elif arg in base_classes.FIELDTYPE:
                # return units on results dict item, ex: stress, strain etc
                ftype = base_classes.FIELDTYPE[arg]
                if ftype in self.units:
                    mystr = ' ('+ self.units[ftype] + ')'
            res.append(mystr)
        return res

    def set_time(self, time):
        """Sets the time in the FeaModel (preprocessor).

        This time is used when setting loads.

        Args:
            time (float): the time to set
        """
        self.time = time

    def get_item(self, item):
        """Returns an item given a string identifying the item.

        Args:
            item (str): 'A0' or 'P0' or 'L0' etc.
        """
        if item[0] == 'P':
            # get point
            items = self.points
            num = int(item[1:])
            res = [a for a in items if a.id == num]
            return res[0]
        elif item[0] == 'L' or item[1] == 'L':
            # get line
            items = self.signlines
            num = int(item[1:])
            items = [a for a in items if a.get_name() == item]
            return items[0]
        elif item[0] == 'A':
            # get area
            items = self.areas
            num = int(item[1:])
            items = [a for a in items if a.id == num]
            return items[0]
        elif item[0] == 'E':
            # get element
            items = self.elements
            items = [a for a in items if a.get_name() == item]
            return items[0]
        elif item[0] == 'N':
            # get node
            items = self.nodes
            items = [a for a in items if a.get_name() == item]
            return items[0]
        else:
            print('Unknown item! Please pass the name of a point, line or area!')

    def get_items(self, items):
        """Returns a list of correctly typed items.

        Input can be a single string or item, or a list of strings identifying
        items.
        """
        res_items = base_classes.listify(items)
        for ind, item in enumerate(res_items):
            if isinstance(item, str):
                res_items[ind] = self.get_item(item)
        return res_items

    def make_matl(self, name):
        """Makes and returns a new material.

        Args:
            name (str): the material's name
        """
        mat = material.Material(name)
        self.matls.append(mat)
        return mat

    def make_part(self):
        """Makes and returns a new part."""
        #p = part.Part(self)
        #self.parts.append(p)
        return partmodule.Part(self)

    def make_problem(self, problem_type='struct', parts='all'):
        """Makes and returns a new problem, which can be solved.

        Args:
            problem_type (str): problem type, options:
                'struct': structural
            parts (str Part or list of Part): Parts the model will analyze.

                Options:
                    - 'all': add all parts. This is the default.
                    - Part: add the single part
                    - list of Part: add these parts
        """
        if parts == 'all':
            parts = self.parts
        prob = problem.Problem(self, problem_type, parts)
        return prob

    def print_summary(self):
        """Prints a summary of the items in the database.
        """
        items = ['parts', 'areas', 'signlines', 'points', 'elements', 'faces',
                 'nodes']
        names = ['parts', 'areas', 'lines', 'points', 'elements', 'faces',
                 'nodes']

        spacer = '----------------------------------'
        print(spacer)
        print("Items in the database:")
        for (item_str, name) in zip(items, names):
            items = getattr(self, item_str)
            print(' %s: %i' % (name, len(items)))
        print(spacer)

    def plot_nodes(self, fname='', display=True, title='Nodes', nnum=False):
        """Plots the selected nodes.

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown
            title (str): the plot's title
            nnum (bool): if True node numbers are plotted
        """
        nodes = self.view.nodes
        if len(nodes) > 0:
            # plotting elements
            fig = plt.figure()
            axis = fig.add_subplot(111)

            # plot nodes, this is quicker than individual plotting
            axs = [node.y for node in nodes]
            rads = [node.x for node in nodes]
            axis.scatter(axs, rads, s=7, color='black')
            if nnum:
                for node in nodes:
                    node.label(axis)

            # set units
            [d_unit] = self.get_units('dist')
            plt.title(title)
            plt.xlabel('axial, y'+d_unit)
            plt.ylabel('radial, x'+d_unit)
            plt.axis('scaled')

            # extract max and min for plot window
            radials = [n.x for n in nodes]
            axials = [n.y for n in nodes]

            # finish pot
            base_classes.plot_set_bounds(plt, axials, radials)
            base_classes.plot_finish(plt, fname, display)

        else:
            # no elements exist or no elemnts are selected
            res = ''
            if len(self.nodes) == 0:
                res = 'No nodes exist! Try meshing your parts!'
            else:
                res = 'No nodes are selected! Select some!'
            print(res)

    def plot_elements(self, fname='', display=True, title='Elements',
                      enum=False, nshow=False, nnum=False):
        """Plots the selected elements.

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown
            title (str): the plot's title
            enum (bool): if True element numbers are plotted
            nshow (bool): True=plot nodes, False=don't plot them
            nnum (bool): if True node numbers are plotted
        """
        nodes = self.view.nodes
        elements = self.view.elements
        if len(elements) > 0:
            # plotting elements
            fig = plt.figure()
            axis = fig.add_subplot(111)
            polys = []
            for element in elements:
                poly = element.get_poly()
                polys.append(poly)
            coll = PatchCollection(polys, facecolors=FCOLOR, edgecolors=ECOLOR)
            axis.add_collection(coll)

            # plot element numbers
            if enum:
                for element in elements:
                    element.label(axis)

            # plot nodes, this is quicker than individual plotting
            if nshow:
                axs = [node.y for node in nodes]
                rads = [node.x for node in nodes]
                axis.scatter(axs, rads, s=7, color='black')
                if nnum:
                    for node in nodes:
                        node.label(axis)

            # set units
            [d_unit] = self.get_units('dist')
            plt.title(title)
            plt.xlabel('axial, y'+d_unit)
            plt.ylabel('radial, x'+d_unit)
            plt.axis('scaled')

            # extract nodes for elements if no nodes are selected
            if len(nodes) == 0:
                tmp = set()
                for element in elements:
                    tmp.update(element.nodes)
                nodes = list(tmp)
            # extract max and min for plot window
            radials = [n.x for n in nodes]
            axials = [n.y for n in nodes]

            # finish pot
            base_classes.plot_set_bounds(plt, axials, radials)
            base_classes.plot_finish(plt, fname, display)

        else:
            # no elements exist or no elemnts are selected
            res = ''
            if len(self.elements) == 0:
                res = 'No elements exist! Try meshing your parts!'
            else:
                res = 'No elements are selected! Select some!'
            print(res)

    def plot_pressures(self, fname='', display=True):
        """Plots the load step pressures on the passed part(s) or selected
        elements.

        This is an element plot, with arrows showing pressure magnitude and
        directions.

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown. Default = True
        """
        elements = self.view.elements
        faces = self.view.faces
        nodes = self.view.nodes

        # plot all elements and store length, length determins min pressure arrow
        if len(elements) > 0:
            # plotting elements
            fig = plt.figure()
            axis = fig.add_subplot(111, aspect='equal')

            # plot polys and calculate avg face length
            polys = []
            face_len = []
            for element in elements:
                face_len.append(element.face[1].length())
                poly = element.get_poly()
                polys.append(poly)
            coll = PatchCollection(polys, edgecolors=ECOLOR, facecolors=FCOLOR)
            axis.add_collection(coll)
            face_len = sum(face_len)/len(face_len)

            # store pressures we'll want to plot: list of [face, pval]
            plist = []
            for load in self.loads[self.time]:
                if load.ltype in  ['press', 'press_fluid']:
                    loadlist = load.get_list()
                    for [face, pval] in loadlist:
                        if face in faces:
                            plist.append([face, pval])
            pressures = [abs(pval) for [face, pval] in plist]

            # set max and min pressure bounds
            pmin = min(pressures)
            pmax = max(pressures)
            # extract nodes for elements if no nodes are selected
            if len(nodes) == 0:
                tmp = set()
                for element in elements:
                    tmp.update(element.nodes)
                nodes = list(tmp)
            radials = [p.x for p in nodes]
            axials = [p.y for p in nodes]
            # arrow length = arrow_min + (pval-pmin)*mult
            arrow_min = face_len
            mult = 0
            if pmax != pmin:
                adelta = max(axials) - min(axials)
                rdelta = max(radials) - min(radials)
                delta = max(adelta, rdelta)
                dist = delta*0.2
                mult = dist/(pmax - pmin)

            # make tick list for later plot, and color map
            cbar_val = None
            tick_list = []
            cmap = None
            if pmax != pmin:
                # we have a range of values we're plotting
                tick_list = linspace(pmin, pmax, 8)
                cmap = plt.get_cmap(CMAP)
            else:
                cbar_val = pmin
                pmax = pmin + 1.0
                pmin = pmin - 1.0
                tick_list = [pmin, cbar_val, pmax]  # default 3 values to plot one solid color
                cmap = colors.ListedColormap(['b', 'b']) # default to plot one val

            # set color contours for arrows
            cnorm = colors.Normalize(vmin=pmin, vmax=pmax)
            scalarmap = cmx.ScalarMappable(norm=cnorm, cmap=cmap)
            scalarmap.set_array([])

            # make arrows, store axials + radials
            for [face, pval] in plist:
                [face_point, unit] = face.get_mnorm()
                arrow_length = arrow_min + (abs(pval) - pmin)*mult
                other_point = face_point + unit*arrow_length
                radials.append(other_point.x)
                axials.append(other_point.y)

                # compression
                tail = other_point
                delta = face_point - other_point
                if pval < 0:
                    # tension
                    tail = face_point
                    delta = other_point - face_point

                headwidth = face_len*0.2
                headlength = face_len*0.3
                colorval = scalarmap.to_rgba(abs(pval))
                plt.arrow(tail.y, tail.x, delta.y, delta.x,
                          color=colorval,
                          head_width=headwidth, head_length=headlength,
                          length_includes_head=True)

            # set the horizontal and vertical axes
            base_classes.plot_set_bounds(plt, axials, radials)

            # set units and titles
            [d_unit, p_unit, t_unit] = self.get_units('dist', 'stress', 'time')
            tstr = 'Pressures %s\nTime=%f%s' % (p_unit, self.time, t_unit)
            plt.title(tstr)
            plt.xlabel('axial, y'+d_unit)
            plt.ylabel('radial, x'+d_unit)

            # set the colorbar
            cbar = plt.colorbar(scalarmap, orientation='vertical',
                                ticks=tick_list)
            if cbar_val != None:
                cbar.ax.set_yticklabels(['', str(cbar_val), ''])
            base_classes.plot_finish(plt, fname, display)

        else:
            res = ''
            if len(self.elements) == 0:
                res = 'No elements exist! Try meshing your parts!'
            else:
                res = 'No elements are selected! Select some!'
            print(res)

    def plot_constraints(self, fname='', display=True):
        """Plots the constraints on the passed part(s) or selected items.

        This is an element and node plot, with arrows showing displacement
        constraints.

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown. Default = True
        """
        elements = self.view.elements
        nodes = self.view.nodes

        # plot all elements
        if len(elements) > 0:
            # plotting elements
            fig = plt.figure()
            axis = fig.add_subplot(111, aspect='equal')

            # plot polys and calculate avg face length
            polys = []
            face_len = []
            for element in elements:
                face_len.append(element.face[1].length())
                poly = element.get_poly()
                polys.append(poly)
            coll = PatchCollection(polys, edgecolors=ECOLOR, facecolors=FCOLOR)
            axis.add_collection(coll)
            face_len = sum(face_len)/len(face_len)

            # store displacements we'll plot: list of [node, dict ux,uy,uz]
            ulist = []
            vals = []
            for load in self.loads[self.time]:
                if load.ltype in  ['ux', 'uy', 'uz']:
                    alist = load.get_list()
                    for nodelist in alist:
                        node = nodelist[0]
                        if node in nodes:
                            ulist += [nodelist]
                            vals += list(nodelist[1].values())

            # check min and max bounds
            pmin = min(vals)
            pmax = max(vals)

            # make tick list for later plot, and color map
            cbar_val = None
            tick_list = []
            cmap = None
            if pmax != pmin:
                # we have a range of values we're plotting
                tick_list = linspace(pmin, pmax, 8)
                cmap = plt.get_cmap(CMAP)
            else:
                cbar_val = pmin
                pmax = pmin + 1.0
                pmin = pmin - 1.0
                tick_list = [pmin, cbar_val, pmax]  # default 3 values to plot one solid color
                cmap = colors.ListedColormap(['b', 'b']) # default to plot one val

            # set color contours for arrows
            cnorm = colors.Normalize(vmin=pmin, vmax=pmax)
            scalarmap = cmx.ScalarMappable(norm=cnorm, cmap=cmap)
            scalarmap.set_array([])

            # make arrows for displacements
            # extract nodes for elements if no nodes are selected
            if len(nodes) == 0:
                tmp = set()
                for element in elements:
                    tmp.update(element.nodes)
                nodes = list(tmp)
            radials = [p.x for p in nodes]
            axials = [p.y for p in nodes]
            pvect = {'ux':geometry.Point(1, 0, 0),
                     'uy':geometry.Point(0, 1, 0),
                     'uz':geometry.Point(0, 0, 1)}
            for [node, udict] in ulist:
                for (key, val) in udict.items():
                    headw = face_len*0.4
                    headl = headw
                    point = geometry.Point(node.x, node.y, node.z)
                    unit = pvect[key]

                    # nonzero displacement, draw from node
                    tail = point
                    head = tail + unit*val
                    store_point = head
                    if val == 0:
                        # zero displ, draw to node
                        head = point
                        tail = head - unit*face_len
                        store_point = tail
                    delta = head - tail
                    radials.append(store_point.x)
                    axials.append(store_point.y)

                    colorVal = scalarmap.to_rgba(val)
                    plt.arrow(tail.y, tail.x, delta.y, delta.x,
                              color=colorVal, head_width=headw,
                              head_length=headl, length_includes_head=True)

            # set the horizontal and vertical axes
            base_classes.plot_set_bounds(plt, axials, radials)

            # set units + titles
            [d_unit, t_unit] = self.get_units('dist', 'time')
            tstr = 'Constraints %s\nTime=%f%s' % (d_unit, self.time, t_unit)
            plt.title(tstr)
            plt.xlabel('axial, y'+d_unit)
            plt.ylabel('radial, x'+d_unit)

            # set the colorbar
            cbar = plt.colorbar(scalarmap, orientation='vertical',
                                ticks=tick_list)
            if cbar_val != None:
                cbar.ax.set_yticklabels(['', str(cbar_val), ''])
            base_classes.plot_finish(plt, fname, display)

        else:
            res = ''
            if len(self.elements) == 0:
                res = 'No elements exist! Try meshing your parts!'
            else:
                res = 'No elements are selected! Select some!'
            print(res)

    def plot_multiple(self, fname='', display=True, title='',
                      styledict={'labels':['points', 'lines', 'areas', 'parts'],
                                 'items':['points, lines, areas']}):
        """Plots items of type styledict['items'], labels styledict['labels']

        Only the items currently selected will be plotted.
        """
        # http://stackoverflow.com/questions/470690/how-to-automatically-generate-n-distinct-colors/4382138#4382138
        # http://stackoverflow.com/questions/2328339/how-to-generate-n-different-colors-for-any-natural-number-n
        # start plotting
        fig = plt.figure()
        axis = fig.add_subplot(111)

        # check if we're plotting parts or areas and pick a color map
        colormaker, ids = None, None
        color_plot = False
        if 'parts' in styledict['items']:
            color_plot = True
            ids = [item.id for item in self.view.parts]
            if 'areas' in styledict['items']:
                styledict['items'].remove('areas')
        if 'areas' in styledict['items']:
            color_plot = True
            ids = [item.id for item in self.view.areas]
        if color_plot:
            cmap = plt.get_cmap(GEOM_CMAP)
            norm = colors.Normalize(vmin=min(ids), vmax=max(ids))
            colormaker = cmx.ScalarMappable(norm=norm, cmap=cmap)

        # plot the items
        for item_type in ['points', 'lines', 'areas', 'parts']:
            plot_on = item_type in styledict['items']
            label_on = item_type in styledict['labels']
            items = getattr(self.view, item_type)
            if plot_on and label_on:
                for item in items:
                    if color_plot and item_type in ['areas', 'parts']:
                        color = colormaker.to_rgba(item.id)
                        item.plot(axis, True, color)
                    else:
                        item.plot(axis, True)
            elif plot_on:
                for item in items:
                    if color_plot and item_type in ['areas', 'parts']:
                        color = colormaker.to_rgba(item.id)
                        item.plot(axis, False, color)
                    else:
                        item.plot(axis, False)
            elif label_on:
                for item in items:
                    item.label(axis)

        # set the horizontal and vertical axes
        points = self.view.points
        radials = [point.x for point in points]
        axials = [point.y for point in points]
        base_classes.plot_set_bounds(plt, axials, radials)

        # set units
        [d_unit] = self.get_units('dist')

        # show plot
        plt.title(title)
        plt.xlabel('axial, y'+d_unit)
        plt.ylabel('radial, x'+d_unit)
        axis.set_aspect('equal')
        base_classes.plot_finish(plt, fname, display)

    def plot_parts(self, fname='', display=True, title='Parts', label=True):
        """Plots selected parts

        Defaults to displaying labels on: parts and filling all part areas.

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown. Default = True
            title (str): the plot's title
            label (bool): if True all Part objects will be labeled
        """
        item = 'parts'
        styledict = {'items':[], 'labels':[]}
        if label:
            styledict['items'].append(item)
            styledict['labels'].append(item)
        else:
            styledict['items'].append(item)

        self.plot_multiple(fname, display, title, styledict)

    def plot_areas(self, fname='', display=True, title='Areas', label=True):
        """Plots selected areas

        Defaults to displaying labels on: areas and filling all part areas.

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown. Default = True
            title (str): the plot's title
            label (bool): if True all areas will be labeled
        """
        item = 'areas'
        styledict = {'items':[], 'labels':[]}
        if label:
            styledict['items'].append(item)
            styledict['labels'].append(item)
        else:
            styledict['items'].append(item)

        self.plot_multiple(fname, display, title, styledict)

    def plot_lines(self, fname='', display=True, title='Lines', label=True):
        """Plots selected lines and arcs

        Defaults to displaying labels on: lines and arcs

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown. Default = True
            title (str): the plot's title
            label (bool): if True all lines and arcs will be labeled
        """
        item = 'lines'
        styledict = {'items':[], 'labels':[]}
        if label:
            styledict['items'].append(item)
            styledict['labels'].append(item)
        else:
            styledict['items'].append(item)

        self.plot_multiple(fname, display, title, styledict)

    def plot_points(self, fname='', display=True, title='Points', label=True):
        """Plots selected points

        Defaults to displaying labels on: points

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown. Default = True
            title (str): the plot's title
            label (bool): if True all points will be labeled
        """
        item = 'points'
        styledict = {'items':[], 'labels':[]}
        if label:
            styledict['items'].append(item)
            styledict['labels'].append(item)
        else:
            styledict['items'].append(item)

        self.plot_multiple(fname, display, title, styledict)

    def plot_geometry(self, fname='', display=True, title='Geometry', pnum=True,
                      lnum=True, anum=True, afill=True):
        """Plots selected geometry items: areas, lines, points.

        Defaults to displaying labels on: areas, lines, and points, and filling
        in the areas.

        Args:
            fname (str): png image file prefix, if given, image will be saved
            display (bool): if True, interactive plot will be shown, if False
                plot will not be shown. Default = True
            title (str): the plot's title
            pnum (bool): if True all Point objects will be labeled
            lnum (bool): if True all Line and Arc objects will be labeled
            anum (bool): if True all Area objects will be labeled
            afill (bool): if True all Area objects will be filled
        """
        styledict = {'items':[], 'labels':[]}
        if pnum:
            styledict['items'].append('points')
            styledict['labels'].append('points')
        if lnum:
            styledict['items'].append('lines')
            styledict['labels'].append('lines')
        if anum:
            styledict['labels'].append('areas')
            if afill:
                styledict['items'].append('areas')
        elif anum == False and afill:
            styledict['items'].append('areas')

        self.plot_multiple(fname, display, title, styledict)

    @staticmethod
    def __get_cname(items):
        """Returns a component name prefix, for labeling lists of items.

        Args:
            items (list): a list of the items we'll be making a component of
        """
        cname = ''
        if len(items) == 1:
            if items[0] == 'all':
                cname = 'all'
            else:
                cname = items[0].get_name()
        else:
            cname = items[0].get_name()+'-'+items[-1].get_name()
        return cname

    def __get_make_comp(self, comp):
        """Stores component if it doesn't exist, returns it if it does.

        Args:
            comp (Component): component we want to store/get
        """
        if comp not in self.components:
            comp = self.components.append(comp)
        else:
            ind = self.components.index(comp)
            comp = self.components[ind]
        return comp

    def __get_make_surfint(self, surfint):
        """Stores surfac interaction if it doesn't exist, returns it if it does.

        Args:
            surfint (connectors.SurfaceInteraction): item to get or make
        """
        items = [item for item in self.surfints if item.int_type == surfint.int_type]
        if surfint.int_type == 'LINEAR':
            for item in items:
                if item.k == surfint.k:
                    return item
        elif surfint.type == 'EXPONENTIAL':
            for item in items:
                if item.c0 == surfint.c0 and item.p0 == surfint.p0:
                    return item
        surfint = self.surfints.append(surfint)
        return surfint

    def register(self, item):
        """Adds an item to the feamodel.

        Item is added to the correct list and its id is updated.
        Allowed Items:

        * Point
        * Line
        * Arc
        * SignLine
        * SignArc
        * LineLoop
        * Area
        * Part
        """
        class_name = item.__class__.__name__
        class_to_list = {'Point': 'points',
                         'Line': 'lines',
                         'Arc': 'lines',
                         'SignLine': 'signlines',
                         'SignArc': 'signlines',
                         'LineLoop': 'lineloops',
                         'Area': 'areas',
                         'Part': 'parts'}
        if class_name in class_to_list:
            list_name = class_to_list[class_name]
            getattr(self, list_name).append(item)
        else:
            print('ERROR: the item you passed must be a geometry item!')
            print('You passed a %s' % class_name)
            message = ("Allowed items: Point, Line, Arc, SignLine, SignArc, "
                       "LineLoop, Area, Part")
            print(message)

    def __add_load(self, load, time):
        """Adds a load to the FeaModel.

        Args:
            load (Load or Load_Linear): th eload to add
            time (float): the load's time point. The first time point is 1.0
        """
        if time in self.loads:
            self.loads[time].append(load)
        else:
            self.loads[time] = [load]
        return load

    def scale(self, unitstr, point_first=None, point_last=None):
        """Scales the points from [point_first, ..., point_last] using unitstr

        If point_first and point_last are both None, all points are scaled.

        Args:
            unitstr (str): string scalar 'fromunit-tounit' using the below units:

                * mm, m, in, ft
                * Examples 'mm-m' 'm-in' 'ft-mm' 'in-ft'
                * Default value is '' and does not apply a scale factor

            point_first (Point or None or str): the first point to scale,
                string point names may be passed
            point_last (Point or None or str): the last point to scale,
                string point names may be passed
        """
        # convert to points if passing in string point names 'P0'
        if isinstance(point_first, str):
            point_first = self.get_item(point_first)
        if isinstance(point_last, str):
            point_last = self.get_item(point_last)

        ind_first = 0
        if point_first != None:
            ind_first = self.points.index(point_first)
        ind_last = -1
        if point_last != None:
            ind_last = self.points.index(point_last)

        unit_size = {'m':1.0, 'mm':.001, 'ft':.3048, 'in':0.0254}
        units = unitstr.split('-')
        from_unit, to_unit = units[0], units[1]
        scalar = unit_size[from_unit]/unit_size[to_unit]
        # scale all points in set
        # update all dependent arcs
        # update all dependent line loops
        # update all dependent areas
        # update all dependent parts

    def set_gravity(self, grav, items):
        """Sets gravity on the elements in items.

        Assumes gravity acts in the -x direction with magnitude grav.

        Args:
            grav (float): gravity acceleration, MUST BE POSTIVE
            items (Area or Part or list): items gravity acts on

                - Area: gravity acts on elements in this area
                - Part: gravity acts on elements in this part
                - list of Part or Area: gravity acts on their elements
        """
        items = self.get_items(items)
        ctype = 'elements'
        cname = self.__get_cname(items)

        # make compoenet
        comp = components.Component(items, ctype, cname)
        comp = self.__get_make_comp(comp)

        # make load
        ltype = 'gravity'
        time = self.time
        load = loads.ConstLoad(ltype, comp, grav)
        self.__add_load(load, time)
        return load

    def set_rpm(self, rpm, items):
        """Sets rpm rotation load on the items.

        Args:
            rpm (float): rotation per minute
            items (Area or Part or list): items rotation acts on

                - Area: rotation acts on elements in this area
                - Part: rotation acts on elements in this part
                - list of Part or Area: rotation acts on their elements
        """
        # applies rpm to the items
        items = self.get_items(items)
        ctype = 'elements'
        cname = self.__get_cname(items)

        # make compoenet
        comp = components.Component(items, ctype, cname)
        comp = self.__get_make_comp(comp)

        # make load
        ltype = 'rpm'
        time = self.time
        load = loads.ConstLoad(ltype, comp, rpm)
        self.__add_load(load, time)
        return load

    def set_radps(self, radps, items):
        """Sets radians per second rotation load on the items.

        Args:
            radps (float): radians per second
            items (Area or Part or list): items rotation acts on

                - Area: rotation acts on elements in this area
                - Part: rotation acts on elements in this part
                - list of Part or Area: rotation acts on their elements
        """
        items = self.get_items(items)
        ctype = 'elements'
        cname = self.__get_cname(items)

        # make compoenet
        comp = components.Component(items, ctype, cname)
        comp = self.__get_make_comp(comp)

        # make load
        ltype = 'radps'
        time = self.time
        load = loads.ConstLoad(ltype, comp, radps)
        self.__add_load(load, time)
        return load

    def set_fluid_press(self, items, rho, g, xo, po):
        """This sets a fluid presure load on items.

        This fluid pressure is dependednt on the x axis.
        g must be positive.

            - P = f(x)
            - P(xo) = po
            - P(x) = po + rho*g*(xo - x)

        Args:
            items (str or Line or Arc or list): items to set pressure on

                - str: string name of Line or Arc item, for example 'L0'
                - Line or Arc: set presssure on this
                - list or Line or Arc: set pressure on these
            rho (float): fluid density in mass/volume
            g (+float): gravity in dist/(t^2), MUST BE POSITIVE
            xo (float): sea level height, MUST BE POSITIVE
            po (float): sea level pressure, MUST BE POSITIVE
        """
        items = self.get_items(items)
        ctype = 'faces'
        cname = self.__get_cname(items)

        # make compoenet
        comp = components.Component(items, ctype, cname)
        comp = self.__get_make_comp(comp)

        # make load
        ltype = 'press_fluid'
        mult = rho*g
        load = loads.LinearLoad(ltype, comp, po, mult, xo)
        self.__add_load(load, self.time)
        return load

    def set_load(self, ltype, items, lval, ldir=None):
        """Sets a pressure or force load on line(s).

        The force load is divided by the number of nodes and applied on
        each node.

        The pressure load is applied to the child faces under the line.
        Positive is compression, negative is tension.

        Args:
            ltype (str): 'press' or 'force'
            items (str or Line or Arc or list): items to set load on

                - str: string name of Line or Arc item, for example 'L0'
                - Line or Arc: set load on this
                - list or Line or Arc: set load on these
            lval (float): load value.

                - For ltype = 'press' this is in force/area units
                - For ltype = 'force' this is in force units
            ldir (None or str): load direction. Defaults to None

                - str: when ltype='load', we need to set ldir to 'x' or 'y'
                - None: when ltype='press'
        """
        items = self.get_items(items)
        ctype = 'nodes'
        if ltype == 'press':
            ctype = 'faces'
        cname = self.__get_cname(items)

        # make compoenet
        comp = components.Component(items, ctype, cname)
        comp = self.__get_make_comp(comp)

        # make load
        if ltype == 'force':
            ltype = 'f'+ldir # for example fx
        load = loads.ConstLoad(ltype, comp, lval)
        self.__add_load(load, self.time)
        return load

    def set_constr(self, ltype, items, axis, val=0.0):
        """Sets a displacement constraint on the passed item(s).

        Args:
            ltype (str): 'fix' or 'displ'

                - 'fix': val arg should not be passed
                - 'displ': val arg must be passed
            items (str, SignLine, SignArc, Point or list): item(s) to apply the
                constraint on

                - str: this string must be a line or point name
                - SignLine or SignArc: the constraint will be applied
                    to this item
                - Point: the constraint will be applied on this item
                - list of SignLine or SignArc: the constraint will be appplied
                    on these
                - list of Point: the constraint will be appplied on these
                - list of str names: pass line or point names, constr
                    applied on them

            axis (str): load axis, 'x' or 'y'
            val (float): displacement value, defaults to 0.0, only needs to be
                set when ltype = 'displ'
        """
        items = self.get_items(items)
        cname = self.__get_cname(items)
        ctype = 'nodes'

        # make compoenet
        comp = components.Component(items, ctype, cname)
        comp = self.__get_make_comp(comp)

        # make load
        ltype = 'u'+axis # for example ux
        load = loads.ConstLoad(ltype, comp, val)
        self.__add_load(load, self.time)
        return load

    def set_contact_linear(self, master_lines, slave_lines, kval, many_si=False):
        """Sets contact between master and slave lines.

        Slave lines are on the more compliant or more curved object.

        Args:
            master_lines (list): list of SignLine or SignArc
            slave_lines (list): list of SignLine or SignArc
            kval (float): stiffness, 5 to 50 times the youngs modulus
                of the touching matl
            many_si (bool): True, make unique surface interaction for every contact
                False, use existing surface interaction if we can
        """

        master_items = self.get_items(master_lines)
        master_cname = self.__get_cname(master_items)
        ctype = 'faces'
        master_comp = components.Component(master_items, ctype, master_cname)
        master_comp = self.__get_make_comp(master_comp)

        slave_items = self.get_items(slave_lines)
        slave_cname = self.__get_cname(slave_items)
        ctype = 'faces'
        slave_comp = components.Component(slave_items, ctype, slave_cname)
        slave_comp = self.__get_make_comp(slave_comp)

        surf_int = connectors.SurfaceInteraction('LINEAR', kval)
        if many_si:
            surf_int = self.surfints.append(surf_int)
        else:
            surf_int = self.__get_make_surfint(surf_int)

        cont = connectors.Contact(master_comp, slave_comp, surf_int, True)
        self.contacts.append(cont)


    def set_eshape(self, eshape='quad', eorder=2):
        """Sets the element shape and order to use when meshing the model.

        Args:
            eshape (str): element shape

                - 'quad': quadrilaterials
                - 'tri': triangles
            eorder (int): element order, default=2

                - 1: corner nodes only (3 and 4 noded elements)
                - 2: corder nodes and midside nodes (6 and 8 noded elements)
        """
        self.eshape = eshape # quad or tri
        self.eorder = eorder # 1 or 2

    def set_etype(self, etype, items, thick=None):
        """Sets the element type, and thickness on areas or parts.

        Args:
            etype (str): element type

                - 'plstress': plane stress
                - 'plstrain': plane strain
                - 'axisym': axisymmetric
            items (str or Area or Part or list): set element type on these

                - str: string name of Area or Part item. Example: 'A0', 'PART0'
                - Area: set element type on the elements in this area
                - Part: set element type on the elements in this part
                - list of Area or Part: set element type on their elements
            thick (float or None): element thickness

                - None: default, used for axisymmetric element type
                - float: thickness value to use for plane stress or plane
                    strain elements
        """
        items = self.get_items(items)

        # set the element types on the areas, this is used to fix
        # elements when importing them from the inp file
        for item in items:
            if isinstance(item, geometry.Area):
                item.set_etype(etype)
            if isinstance(item, partmodule.Part):
                for area in item.areas:
                    area.set_etype(etype)

        # set a thickness component if needed
        if etype != 'axisym' and thick != None:

            # make component for element thickness
            cname = self.__get_cname(items)
            ctype = 'nodes'
            comp = components.Component(items, ctype, cname)
            comp = self.__get_make_comp(comp)

            # add load
            ltype = 'nodal_thickness'
            time = 0.0
            load = loads.ConstLoad(ltype, comp, thick)
            self.__add_load(load, time)
            return load
        return None

    def set_matl(self, matl, items):
        """Sets the material on Part or Area items.

        Args:
            matl (Material): material to assign to items
            items (Part or Area or list): items that will have matl assigned

                - Part: assign matl to this
                - Area: assign matl to this
                - list of Area or Part: assign matl to these
        """
        items = self.get_items(items)
        cname = self.__get_cname(items)

        # store the material if it's not already
        if matl not in self.matls:
            matl = self.matls.append(matl)

        # make component
        ctype = 'elements'
        comp = components.Component(items, ctype, cname)
        comp = self.__get_make_comp(comp)

        # make load
        ltype = 'matl'
        time = 0.0
        load = loads.ConstLoad(ltype, comp, matl)
        self.__add_load(load, time)
        return load

    def __read_inp(self, fname):
        """Reads in the mesh from a Calculix inp file.

        All nodes, elements, and faces are read. Child nodes, elements and
        faces are assigned to their geometry parents (points, lines, arcs,
        areas, parts.

        Args:
            fname (str): file name to read, must include '.inp' extension
        """
        f = open(fname, 'r')
        mode = None
        set_name = None
        set_type = None

        items = [] # holder for nodes or elements in nsets or esets
        N = base_classes.Meshlist() # store nodes
        E = base_classes.Meshlist() # store elements, allows renumbering before putting int model
        F = [] # store faces
        sets = {'E':{}, 'N':{}} # store sets
        etype = ''
        Dict_NodeIDs={}
        Dict_ElemIDs={}

        # read in input file
        for line in f:
            if line[0] != '*':
                if mode == 'nmake':
                    L = line.split(',')
                    L = [a.strip() for a in L]
                    (nnum, x, y, z) = (int(L[0]), float(L[1]), float(L[2]), float(L[3]))
                    node = mesh.Node(nnum, x, y, z)
                    N.append(node)
                    Dict_NodeIDs[nnum]=node
                elif mode == 'emake':
                    L = line.split(',')
                    L = [int(a.strip()) for a in L]
                    enum = L[0]
                    nlist = [Dict_NodeIDs[a] for a in L[1:]]
                    e = mesh.Element(enum, etype, nlist)
                    faces = e.faces
                    E.append(e)
                    Dict_ElemIDs[enum]=e
                    F += faces
                    sets[set_type][set_name].append(e)
                elif mode == 'set':
                    L = line.split(',')
                    L = [a.strip() for a in L]
                    L = [int(a) for a in L if a != '']
                    items = []
                    if set_type == 'E':
                        items = [Dict_ElemIDs[a] for a in L if a in Dict_ElemIDs.keys()]
                    elif set_type == 'N':
                        items = [Dict_NodeIDs[a] for a in L]
                    if items == [None]*len(items):
                        pass # the elements were not found
                    else:
                        sets[set_type][set_name] += items

            # mode setting
            if '*Node' in line or '*NODE' in line:
                mode = 'nmake'
            elif '*Element' in line or '*ELEMENT' in line:
                L = line.split(',') # split it based on commas
                e = L[1].split('=')
                etype = e[1]

                # exclude T elements made in gmsh
                if etype[0] != 'T':
                    e = L[2].split('=')
                    set_name = e[1].strip()
                    set_type = 'E'
                    sets[set_type][set_name] = []
                    mode = 'emake'
                else:
                    mode = None
            elif '*ELSET' in line:
                L = line.split(',')
                e = L[1].split('=')
                set_name = e[1].strip()
                set_type = 'E'
                sets[set_type][set_name] = []
                mode = 'set'
            elif '*NSET' in line:
                L = line.split(',')
                e = L[1].split('=')
                set_name = e[1].strip()
                set_type = 'N'
                sets[set_type][set_name] = []
                mode = 'set'
        f.close()

        # loop through sets and remove empty sets
        # store sets to delete
        todel = []
        for (set_type, set_dict) in sets.items():
            for (set_name, item_list) in set_dict.items():
                if item_list == []:
                    todel.append({'set_type':set_type, 'set_name':set_name})
        # delete the empty sets
        for adict in todel:
            (set_type, set_name) = (adict['set_type'], adict['set_name'])
            del sets[set_type][set_name]
            #print('Empty set type:%s name:%s deleted' % (set_type, set_name))

        # this resets the min element to number 1
        if E.get_minid() > 1:
            E.set_minid(1)

        #-----------------------------------
        # Node and element assignment back onto parts, areas, lines, points
        #-----------------------------------
        # assign elements + nodes to feamodel
        self.elements = E
        self.faces = F

        # remove arc center nodes from imported node set
        # those nodes have no elements under them
        torem = []
        for node in N:
            if len(node.elements) == 0:
                torem.append(node)
        for node in torem:
            N.remove(node)
        # assign nodes to feamodel
        self.nodes = N

        for part in self.parts:
            # assign part element and node sets
            pname = part.get_name()
            part.elements = sets['E'][pname]
            part.nodes = sets['N'][pname]

            # assign all nodes and elements to areas, fix the element types
            for area in part.areas:
                aname = area.get_name()
                area.elements = sets['E'][aname]
                area.nodes = sets['N'][aname]
                area.set_child_ccxtypes() #paint element types on elements

            # assign the child nodes to points
            pts = part.points
            for point in pts:
                ndist = []
                for node in part.nodes:
                    p_tmp = geometry.Point(node.x, node.y)
                    dist = point - p_tmp
                    dist = dist.length()
                    ndist.append({'dist':dist, 'node':node})
                # sort the list by dist, sorts low to high
                ndist = sorted(ndist, key=lambda k: k['dist'])
                point.nodes = [ndist[0]['node']]
                #print('Point %s = node %s' % (pt, pt.nodes))

            # assign the nodes and n1 and faces to lines
            slines = part.signlines
            for sline in slines:
                lname = sline.get_name()
                area = sline.lineloop.parent
                nodes = sets['N'][sline.line.get_name()]
                n1 = [n for n in nodes if n.order == 1]
                sline.nodes = nodes
                sline.n1 = n1
                # make a set of all faces that contain the line's node
                allfaces = set()
                for node in nodes:
                    allfaces.update(node.faces)
                # make a set of all faces on the line only
                linefaces = set()
                for face in allfaces:
                    if set(face.nodes).issubset(set(nodes)):
                        linefaces.add(face)
                # and only the ones whose element centers are in the area
                faces = set()
                for face in linefaces:
                    is_lineface = area.contains_point(face.element.center)
                    if is_lineface:
                        faces.add(face)
                sline.faces = list(faces)
        print('Elements: %i' % len(E))
        print('Nodes: %i' % len(N))
        print('Done reading Calculix/Abaqus .inp file')

    def mesh(self, size=1.0, meshmode='fineness', mesher='gmsh'):
        """Meshes all parts.

        Args:
            size (float):

                - if meshmode == 'fineness' (default):
                    - mesh size is adapted to geometry size
                    - set size = 0.0001 - 1.0, to define how fine the mesh is.
                    - Low numbers are very fine, higher numbers are coarser.

                - if meshmode == 'esize':
                    - element size is kept constant
                    - choose it depending on geometry size
                    - it should be reduced e.g. at arcs with small radius, by calling line.esize function

            meshmode (str):

                - 'fineness': adapt mesh size to geometry
                - 'esize': keep explicitly defined element size

                meshmode is changed to 'esize' is used if esize property is set to points or lines

            mesher (str): the mesher to use

                - 'gmsh': mesh with Gmsh, this is reccomended, it allows holes
                - 'cgx': mesh with Calculix cgx, it doesn't allow holes
        """


        #check if element size is set to points and change meshmode if necessary
        for pt in self.points:
            if pt.esize != None:
                if meshmode=='fineness': print('meshmode is changed to esize, because elementsize was defined on points!')
                meshmode = 'esize'


        #if meshmode esize is chosen: ediv's on lines and arcs are transformed to element sizes on start and end point
        if meshmode == 'esize':
            for line in self.lines:
                if line.ediv != None:
                    line.pt(0).set_esize(line.length()/line.ediv)
                    line.pt(1).set_esize(line.length()/line.ediv)


        if mesher == 'gmsh':
            self.__mesh_gmsh(size, meshmode)
        elif mesher == 'cgx':
            self.__mesh_cgx(size)

    def __mesh_gmsh(self, size, meshmode, timeout=20):
        """Meshes all parts using the Gmsh mesher.

        Args:

            size (float):

                - if meshmode == 'fineness' (default):
                    - mesh size is adapted to geometry size
                    - set size = 0.0001 - 1.0, to define how fine the mesh is.
                    - Low numbers are very fine, higher numbers are coarser.

                - if meshmode == 'esize':
                    - element size is kept constant
                    - choose it depending on geometry size

            meshmode (str):

                - 'fineness': adapt mesh size to geometry
                - 'esize': keep explicitly defined element size

            timeout (int): time in seconds before the process throws a
                subprocess.TimeoutExpired

        """
        geo = []
        ids = {}
        ids['line'] = {}
        ids['plane_surface'] = {}

        # write all points
        for pt in self.points:
            txtline = 'Point(%i) = {%f, %f, %f};' % (pt.id, pt.x, pt.y, 0.0)

            if meshmode == 'esize':
                #add element size to points
                if pt.esize == None:
                    txtline = txtline.replace('}', ', %f}' % (size if self.eshape=='tri' else size*2.))
                    #txtline = txtline.replace('}', ', %f}' % (size))
                else:
                    txtline = txtline.replace('}', ', %f}' % (pt.esize if self.eshape=='tri' else pt.esize*2.))
                    #txtline = txtline.replace('}', ', %f}' % (pt.esize))
            geo.append(txtline)

        # start storing an index number
        ind = self.points[-1].id + 1

        # write all lines
        for line in self.lines:
            lnum = line.id
            ids['line'][lnum] = ind
            pt1 = line.pt(0).id
            pt2 = line.pt(1).id
            txtline = ''
            if isinstance(line, geometry.Arc):
                ctr = line.actr.id
                txtline = 'Circle(%i) = {%i, %i, %i};' % (ind, pt1, ctr, pt2)
            else:
                txtline = 'Line(%i) = {%i,%i};' % (ind, pt1, pt2)
            geo.append(txtline)

            # set division if we have it
            if line.ediv != None and meshmode=='fineness':
                ndiv = line.ediv+1
                esize = line.length()/line.ediv
                if self.eshape == 'quad':
                    ndiv = line.ediv/2+1
                    esize = esize*2
                    # this is needed because quad recombine
                    # splits 1 element into 2
                txtline = 'Transfinite Line{%i} = %i;' % (ind, ndiv)
                print('LINE ELEMENT SIZE: %f, MAKES %i ELEMENTS'
                      % (line.length()/line.ediv, line.ediv))
                geo.append(txtline)
                geo.append('Characteristic Length {%i,%i} = %f;'
                           % (pt1, pt2, esize))
            ind += 1

        # write all areas
        for area in self.areas:
            if area.closed:
                aid = area.id
                aname = area.get_name()
                loop_ids = []
                loops = [area.exlines] + area.holes
                for loop in loops:
                    txtline = 'Line Loop(%i) = ' % (ind)
                    loop_ids.append(str(ind))
                    line_ids = []
                    for sline in loop:
                        lid = ids['line'][sline.line.id]
                        prefix = ''
                        if sline.sign == -1:
                            prefix = '-'
                        line_ids.append('%s%i' % (prefix, lid))
                    txtline = txtline + '{'+ ','.join(line_ids)+'};'
                    geo.append(txtline)
                    ind += 1
                loop_ids = ','.join(loop_ids)
                geo.append('Plane Surface(%i) = {%s};' % (ind, loop_ids))
                ids['plane_surface'][aid] = ind
                geo.append("Physical Surface('%s') = {%i};" % (aname, ind))
                ind += 1

        # write part area components
        for part in self.parts:
            # make components for each part
            txtline = "Physical Surface('%s') = " % (part.get_name())
            area_ids = []
            for area in part.areas:
                if area.closed:
                    aid = ids['plane_surface'][area.id]
                    area_ids.append(str(aid))
            txtline = txtline + '{' + ','.join(area_ids) + '};'
            geo.append(txtline)

        # write all line componenets so we can get nodes out
        for line in self.lines:
            lid = ids['line'][line.id]
            txtline = "Physical Line('%s') = {%i};" % (line.get_name(), lid)
            geo.append(txtline)

        # write node componenets
        # ERROR: node list is not produced by gmsh
        for point in self.points:
            pname = point.get_name()
            txtline = "Physical Point('%s') = {%i};" % (pname, point.id)
            geo.append(txtline)

        # set the meshing options
        if meshmode == 'fineness':
            geo.append('Mesh.CharacteristicLengthFactor = '+str(size)+'; //mesh fineness')
        geo.append('Mesh.RecombinationAlgorithm = 1; //blossom')

        if self.eshape == 'quad':
            geo.append('Mesh.RecombineAll = 1; //turns on quads')
            geo.append('Mesh.SubdivisionAlgorithm = 1; // quadrangles only')
            #geo.append('Mesh.RecombinationAlgorithm = 1; //turns on blossom needed for quad')

        order = self.eorder
        geo.append('Mesh.CharacteristicLengthExtendFromBoundary = 1;')
        geo.append('Mesh.CharacteristicLengthMin = 0;')
        geo.append('Mesh.CharacteristicLengthMax = 1e+022;')
        # use this so small circles are meshed finely
        geo.append('Mesh.CharacteristicLengthFromCurvature = 1;')
        geo.append('Mesh.CharacteristicLengthFromPoints = 1;')
        # geo.append('Mesh.Algorithm = 2; //delauny') #okay for quads
        geo.append('Mesh.Algorithm = 8; //delquad = delauny for quads')
        geo.append('Mesh.ElementOrder = '
                   +str(order)
                   +'; //linear or second set here')
        if order == 2:
            geo.append('Mesh.SecondOrderIncomplete=1; //no face node w/ 2nd order')
        geo.append('Mesh.SaveGroupsOfNodes = 1; // save node groups')

        # write geo file to the local directory
        fname = self.fname+'.geo'
        fout = self.fname+'.inp'
        outfile = open(fname, 'w')
        for line in geo:
            #print (line)
            outfile.write(line+'\n')
        outfile.close()
        print('File: %s was written' % fname)

        # run file in bg mode, -2 is 2d mesh, makes required inp file
        runstr = "%s %s -2 -o %s" % (environment.GMSH, fname, fout)
        print(runstr)
        subprocess.check_call(runstr, timeout=timeout, shell=True)
        print('File: %s was written' % fout)
        print('Meshing done!')

        # write gmsh msh file, for manual checking only
        # not required by pycalculix
        runstr = "%s %s -2 -o %s" % (environment.GMSH, fname,
                                     self.fname+'.msh')
        subprocess.check_call(runstr, timeout=timeout, shell=True)
        print('File: %s.msh was written' % self.fname)

        # read in the calculix mesh
        self.__read_inp(self.fname+'.inp')

    def __mesh_cgx(self, size, meshmode, timeout=20):
        """Meshes all parts using the Calculix cgx mesher.

        Args:

            size (float):

                - if meshmode == 'fineness' (default):
                    - mesh size is adapted to geometry size
                    - set size = 0.0001 - 1.0, to define how fine the mesh is.
                    - Low numbers are very fine, higher numbers are coarser.

                - if meshmode == 'esize':  NOT TESTED WITH CGX
                    - element size is kept constant
                    - choose it depending on geometry size

            meshmode (str):

                - 'fineness': adapt mesh size to geometry
                - 'esize': keep explicitly defined element size  NOT TESTED WITH CGX

            timeout (int): time in seconds before the process throws a
                subprocess.TimeoutExpired
        """
        fbd = []
        comps = []
        cfiles = []

        # Calculix CGX elements
        # axisymmetric
        cgx_elements = {}
        cgx_elements['tri2axisym'] = 'tr6c'
        cgx_elements['tri1axisym'] = 'tr3c'
        cgx_elements['quad2axisym'] = 'qu8c'
        cgx_elements['quad1axisym'] = 'qu4c'
        # plane stress
        cgx_elements['tri2plstress'] = 'tr6s'
        cgx_elements['tri1plstress'] = 'tr3s'
        cgx_elements['quad2plstress'] = 'qu8s'
        cgx_elements['quad1plstress'] = 'qu4s'
        # plane strain
        cgx_elements['tri2plstrain'] = 'tr6e'
        cgx_elements['tri1plstrain'] = 'tr3e'
        cgx_elements['quad2plstrain'] = 'qu8e'
        cgx_elements['quad1plstrain'] = 'qu4e'

        num = 1.0/size
        emult = int(round(num)) # this converts size to mesh multiplier

        # write all points
        for point in self.points:
            linestr = 'pnt %s %f %f %f' % (point.get_name(), point.x, point.y, 0.0)
            fbd.append(linestr)
            # gmsh can't make node componenets so don't do it in cgx
            #L = 'seta %s p %s' % (pt.get_name(), pt.get_name())
            #comps.append(L)

        # write all lines
        for line in self.lines:
            lname = line.get_name()
            pt1 = line.pt(0).get_name()
            pt2 = line.pt(1).get_name()
            linestr = ''
            if isinstance(line, geometry.Arc):
                # line is arc
                pctr = line.actr.get_name()
                linestr = 'line %s %s %s %s' % (lname, pt1, pt2, pctr)
            else:
                # straight line
                linestr = 'line %s %s %s' % (lname, pt1, pt2)
            # set division if we have it
            if line.ediv != None:
                ndiv = self.eorder*line.ediv
                linestr += ' '+str(int(ndiv))
            fbd.append(linestr)
            linestr = 'seta %s l %s' % (lname, lname)
            comps.append(linestr)
            cfiles.append(lname)

        # write all areas
        for area in self.areas:
            if area.closed:
                linestr = 'gsur '+area.get_name()+' + BLEND '
                line_ids = []
                for line in area.signlines:
                    lname = '+ '+line.get_name()
                    if line.sign == -1:
                        lname = '- '+line.get_name()[1:]
                    line_ids.append(lname)
                linestr = linestr + ' '.join(line_ids)
                fbd.append(linestr)
                # add area component, nodes + elements
                astr = 'seta %s s %s' % (area.get_name(), area.get_name())
                fbd.append(astr)
                cfiles.append(area.get_name())

        # write part area components
        for apart in self.parts:
            # make components for each part
            # seta P0 s s0 s1
            line = 'seta %s s ' % apart.get_name()
            cfiles.append(apart.get_name())
            area_ids = []
            for area in apart.areas:
                if area.closed:
                    area_ids.append(area.get_name())
            line = line + ' '.join(area_ids)
            fbd.append(line)

            # mesh all areas
            for area in apart.areas:
                aname = area.get_name()
                estr = self.eshape+str(self.eorder)+area.etype
                cgx_etype = cgx_elements[estr]
                fbd.append('elty %s %s' % (aname, cgx_etype))
            fbd.append('div all mult %i' % emult)
            fbd.append('mesh all')

        # save mesh file
        fbd.append('send all abq')

        # add line and area components
        fbd += comps

        # save component node and element sets
        for comp in cfiles:
            # select nodes under
            fbd.append('comp %s do' % (comp,))
            fbd.append('send %s abq names' % (comp,))

        # this orients the view correctly
        # this is the same as switching to the z+ orientation
        # y us axial, x is radial
        fbd.append('rot z')
        fbd.append('rot c -90')
        fbd.append('plot e all')

        # write fbd file to the local directory
        fname = self.fname+'.fbd'
        f = open(fname, 'w')
        for line in fbd:
            #print (line)
            f.write(line+'\n')
        f.close()
        print('File: %s was written' % fname)

        # run file in bg mode
        runstr = "%s -bg %s" % (environment.CGX, fname)
        p = subprocess.check_call(runstr, timeout=timeout, shell=True)
        print('Meshing done!')

        # assemble the output files into a ccx input file
        inp = []
        files = ['all.msh']
        files += [f+'.nam' for f in cfiles]
        for fname in files:
            infile = open(fname, 'r')
            for line in infile:
                # cgx adds E and N prfixes on sets after =, get rid of these
                if '=' in line and fname != 'all.msh':
                    L = line.split('=')
                    line = L[0] + '=' + L[1][1:]
                    inp.append(line.strip())
                else:
                    inp.append(line.strip())
            infile.close()

            # delete file
            os.remove(fname)

        # write out inp file
        fname = self.fname+'.inp'
        outfile = open(fname, 'w')
        for line in inp:
            #print (line)
            outfile.write(line+'\n')
        outfile.close()
        print('File: %s was written' % fname)

        # read in the calculix mesh
        self.__read_inp(fname)



"""This module sets the dpi and the paths to gmsh, ccx, and cgx.

Attributes:
    DPI (None or float): if high dpi windows 8 monitor value is set, otherwise
        value is None.
    CCX (str): path to Calculix ccx, the sovler.
    CGX (str): path to Calculix cgx, the preprocessor/postprocessor/mesher.
    GMSH (str): path to the Gmsh mesher.
"""

import sys # needed to check if 32 or 64 bit interpreter
import os # used to delete files written by cgx
import platform # need this to check for win and do the below dpi fix
import ctypes # needed to see if it's 32-bit and to get correct win version

# http://stackoverflow.com/questions/19128219/detect-windows-8-1-in-python/22325767#22325767
class OSVERSIONINFOEXW(ctypes.Structure):
    """Returns object w/ attributes that will identify window os version"""
    _fields_ = [('dwOSVersionInfoSize', ctypes.c_ulong),
                ('dwMajorVersion', ctypes.c_ulong),
                ('dwMinorVersion', ctypes.c_ulong),
                ('dwBuildNumber', ctypes.c_ulong),
                ('dwPlatformId', ctypes.c_ulong),
                ('szCSDVersion', ctypes.c_wchar*128),
                ('wServicePackMajor', ctypes.c_ushort),
                ('wServicePackMinor', ctypes.c_ushort),
                ('wSuiteMask', ctypes.c_ushort),
                ('wProductType', ctypes.c_byte),
                ('wReserved', ctypes.c_byte)]

def get_version():
    """Get's the OS version. Returns a float of OS_MAJOR.OS_MINOR
    """
    os_version = OSVERSIONINFOEXW()
    os_version.dwOSVersionInfoSize = ctypes.sizeof(os_version)
    retcode = ctypes.windll.Ntdll.RtlGetVersion(ctypes.byref(os_version))
    if retcode != 0:
        print("Failed to get OS version")
        return -1.0

    major = str(os_version.dwMajorVersion)
    minor = str(os_version.dwMinorVersion)
    return float(major+'.'+minor)

def get_dpi():
    """Returns an int of current DPI for windows or None."""
    if platform.system() == 'Windows':
        version = get_version()
        if version >= 6.2:
            # Windows 8.0 is 6.2.9200
            # Only run on windows 8 or higher to check high dpi monitors

            user32 = ctypes.windll.user32
            w_curr = user32.GetSystemMetrics(0)
            user32.SetProcessDPIAware()
            w_phys = user32.GetSystemMetrics(0)
            curr_dpi = round(w_phys*96/w_curr, 0)

            from pylab import rcParams
            rcParams['figure.dpi'] = curr_dpi
            return curr_dpi
    else:
        return None

def get_paths():
    """Returns a list of paths to: [ccx, cgx, gmsh]."""
    ccx = 'ccx'
    cgx = 'cgx'
    gmsh = 'gmsh'
    return [ccx, cgx, gmsh]

DPI = get_dpi()
CCX, CGX, GMSH = get_paths()



"""This module stores classes that make a finite element analysis mesh.
"""
from matplotlib.patches import Polygon  # needed for plotting elements
from . import geometry

class Element(object):
    """Makes a mesh element.

    Supports 4 noded quads, 8 noded quads, 3 noded tris, 6 noded tris.

    Args:
        enum (int): element id number
        ccxtype (str): ccx element type string, 4 characters long

            Description: element_shape + element_order + element_type

            - element_shape = 'quad' or 'tri'
                - 'quad' = quadrangle
                - 'tri' = triangle
            - element_order = '1' or '2'
                - '1' = corner nodes only (3 or 4 noded element)
                - '2' = corner nodes and midside nodes (6 or 8 noded element)
            - element_type = 'plstress' or 'plstrain' or 'axisym'
                - 'plstress' = plane stress
                - 'plstrain' = plane strain
                - 'axisym' = axisymmetric

            axisymmetric
                - 'CAX6' : 'tri2axisym'
                - 'CAX3' : 'tri1axisym'
                - 'CAX8' : 'quad2axisym'
                - 'CAX4' : 'quad1axisym'
            plane stress
                - 'CPS6' : 'tri2plstress'
                - 'CPS3' : 'tri1plstress'
                - 'CPS8' : 'quad2plstress'
                - 'CPS4' : 'quad1plstress'
            plane strain
                - 'CPE6' : 'tri2plstrain'
                - 'CPE3' : 'tri1plstrain'
                - 'CPE8' : 'quad2plstrain'
                - 'CPE4' : 'quad1plstrain'

        nlist (list): list of Node, the nodes that define the element

    Attributes:
        id (int): element id number
        ccxtype (str): ccx element type string, 4 characters long

            Description: element_shape + element_order + element_type

            - element_shape = 'quad' or 'tri'
                - 'quad' = quadrangle
                - 'tri' = triangle
            - element_order = '1' or '2'
                - '1' = corner nodes only (3 or 4 noded element)
                - '2' = corner nodes and midside nodes (6 or 8 noded element)
            - element_type = 'plstress' or 'plstrain' or 'axisym'
                - 'plstress' = plane stress
                - 'plstrain' = plane strain
                - 'axisym' = axisymmetric

            axisym
                - 'CAX6' : 'tri2axisym'
                - 'CAX3' : 'tri1axisym'
                - 'CAX8' : 'quad2axisym'
                - 'CAX4' : 'quad1axisym'
            plane stress
                - 'CPS6' : 'tri2plstress'
                - 'CPS3' : 'tri1plstress'
                - 'CPS8' : 'quad2plstress'
                - 'CPS4' : 'quad1plstress'
            plane strain
                - 'CPE6' : 'tri2plstrain'
                - 'CPE3' : 'tri1plstrain'
                - 'CPE8' : 'quad2plstrain'
                - 'CPE4' : 'quad1plstrain'

        node (dict): dictionary of nodes, keys are int >= 1
        face (dict): dictionary of faces, keys are int >= 1
        nodes (list): list of element nodes
        faces (list): list of element faces
        center (Point): the center point of the element
    """

    def __init__(self, enum, ccxtype, nlist):
        self.id = enum
        self.ccxtype = ccxtype
        self.node = {}
        self.face = {}

        # calculate the number of faces
        fnum = len(nlist)
        if fnum > 4:
            fnum = int(fnum/2)

        # store nodes
        for (ind, node) in enumerate(nlist):
            if ind >= fnum:
                node.set_order(2)
            node.add_element(self)
            self.node[ind+1] = node

        # store faces
        nodes = nlist[0:fnum]+[nlist[0]]
        for ind in range(fnum):
            node1 = nodes[ind]
            node2 = nodes[ind+1]
            face = Face(ind+1, node1, node2, self)
            if len(nlist) > fnum:
                # we have second order nodes
                face.set_nmid(self.node[ind+1+fnum])
            self.face[ind+1] = face

        # set the element center
        self.center = self.calc_center()

    def __hash__(self):
        """Returns the item's id as its hash."""
        return self.id

    @property
    def faces(self):
        """Returns list of element faces."""
        return self.face.values()

    @property
    def nodes(self):
        """Returns list of element nodes."""
        return self.node.values()

    def get_poly(self):
        """Returns a polygon of the element."""
        nodes = [node for node in self.nodes if node.order == 1]
        # make a list of [[ax, rad],[... ] for making the polygon
        xycoords = [[node.y, node.x] for node in nodes]
        polygon = Polygon(xycoords, closed=True)
        return polygon
    
    def label(self, ax):
        """Labels the element on the passed matplotlib axis."""
        ax.text(self.center.y, self.center.x, self.get_name(),
                ha='center', va='center')

    def calc_center(self):
        """Returns the element center as a Point."""
        pts = [node for node in self.nodes if node.order == 1]
        axials = [point.y for point in pts]
        radials = [point.x for point in pts]
        axial = sum(axials)/len(axials)
        radial = sum(radials)/len(radials)
        return geometry.Point(radial, axial)

    def get_tris(self):
        """Returns a list of triangles for plotting. Triangles are node ids.

        CCX elements are closed in a CCW direction relative to xy  normal
        CCX elements are closed in a CW direction relative to yx mine
        ---> I have to draw in a clockwise direction, otherwise get error
        Triangles are closed in a counter clockwise (CCW) direction to yx mine
        http://matplotlib.org/api/tri_api.html#matplotlib.tri.Triangulation
        """
        res = []
        numnodes = len(self.nodes)
        if numnodes == 3:
            # triangle, first order = 1 triangle
            res.append([self.node[1].id, self.node[3].id, self.node[2].id])
        elif numnodes == 4:
            # quad, first order = 2 triangles
            res.append([self.node[1].id, self.node[3].id, self.node[2].id])
            res.append([self.node[1].id, self.node[4].id, self.node[3].id])
        elif numnodes == 6:
            # tri, second order = 4 triangles
            res.append([self.node[1].id, self.node[6].id, self.node[4].id])
            res.append([self.node[4].id, self.node[5].id, self.node[2].id])
            res.append([self.node[6].id, self.node[3].id, self.node[5].id])
            res.append([self.node[6].id, self.node[5].id, self.node[4].id])
        elif numnodes == 8:
            # quad, second order = 6 triangles
            res.append([self.node[1].id, self.node[8].id, self.node[5].id])
            res.append([self.node[5].id, self.node[6].id, self.node[2].id])
            res.append([self.node[6].id, self.node[7].id, self.node[3].id])
            res.append([self.node[8].id, self.node[4].id, self.node[7].id])
            res.append([self.node[8].id, self.node[7].id, self.node[6].id])
            res.append([self.node[8].id, self.node[6].id, self.node[5].id])
        return res

    def get_area(self):
        """Returns the element area."""
        asum = 0.0
        for face in self.faces:
            node1 = face.nodes[0]
            node2 = face.nodes[1]
            asum += node2.y*node1.x - node2.x*node1.y
        asum = asum*0.5
        return asum

    def set_ccxtype(self, etype):
        """Sets the ccx element type given an etype string.

        Args:
            etype (string): element type string

                - 'plstress' = plane stress
                - 'plstrain' = plane strain
                - 'axisym' = axisymmetric
        """
        fnum = len(self.faces)
        shape = 'tri'
        if fnum == 4:
            shape = 'quad'
        order = '2'
        if len(self.nodes) == fnum:
            order = '1'
        estr = shape+order+etype

        # Calculix CCX elements
        # axisymmetric
        _ccx_elements = {}
        _ccx_elements['tri2axisym'] = 'CAX6'
        _ccx_elements['tri1axisym'] = 'CAX3'
        _ccx_elements['quad2axisym'] = 'CAX8'
        _ccx_elements['quad1axisym'] = 'CAX4'
        # plane stress
        _ccx_elements['tri2plstress'] = 'CPS6'
        _ccx_elements['tri1plstress'] = 'CPS3'
        _ccx_elements['quad2plstress'] = 'CPS8'
        _ccx_elements['quad1plstress'] = 'CPS4'
        # plane strain
        _ccx_elements['tri2plstrain'] = 'CPE6'
        _ccx_elements['tri1plstrain'] = 'CPE3'
        _ccx_elements['quad2plstrain'] = 'CPE8'
        _ccx_elements['quad1plstrain'] = 'CPE4'
        # set element ccxtype
        self.ccxtype = _ccx_elements[estr]

    def ccx(self):
        """Returns text string defining the element for a ccx input file."""
        nids = [str(n.id) for n in self.node.values()]
        val = str(self.id)+', '+', '.join(nids)
        return val

    def get_name(self):
        """Returns the element name based on id number."""
        name = 'E%i' % (self.id)
        return name

    def __str__(self):
        """Returns the element name."""
        nids = [str(node.id) for node in self.nodes]
        return '%s nodes: %s' % (self.get_name(), ','.join(nids))


class Face(object):
    """Makes an element face.

    Args:
        fnum (int): element face number
        node1 (Node): first node
        node2 (Node): second node
        element (Element): parent element

    Attributes:
        id (int): element face number
        nodes (list): list of nodes [node1, node2]
        nmid (Node or None): midside node, defaults to None
        element (Element): parent element
        ext (bool): stores if face is external
    """

    def __init__(self, fnum, node1, node2, element):
        self.id = fnum
        self.nmid = None
        self.ext = False
        self.element = element
        self.nodes = [node1, node2]
        for node in self.nodes:
            node.add_face(self)

    def __hash__(self):
        """Returns a hash = self.element.id*4, Lets us make sets of faces."""
        # we assume the largest number of faces is 4, face id is 1-4, need 0-3
        hval = self.element.id*4+(self.id-1)
        return hval

    def length(self):
        """Return the length of the face."""
        point0 = geometry.Point(self.nodes[0].x, self.nodes[0].y)
        point1 = geometry.Point(self.nodes[1].x, self.nodes[1].y)
        vect = point1 - point0
        return vect.length()

    def get_mnorm(self):
        """Returns a list: [midpoint, normal_vector]

        Returns:
            midpoint (Point): face midpoint
            norm_vect (Point): face normal vector, it is a unit vector
        """
        # return midpt, normal
        point0 = geometry.Point(self.nodes[0].x, self.nodes[0].y)
        point1 = geometry.Point(self.nodes[1].x, self.nodes[1].y)
        midpt = point0 + point1
        midpt = midpt*0.5
        norm_vect = point1 - point0
        norm_vect.rot_ccw_deg(90)
        norm_vect.make_unit()
        return [midpt, norm_vect]

    def set_ext(self):
        """Sets the ext boolean flag to True. Face is external."""
        self.ext = True

    def set_nmid(self, nmid):
        """Sets the face midside node, nmid.

        Args:
            nmid (Point): passed node to set as the face midside node
        """
        self.nmid = nmid
        nmid.add_face(self)
        self.nodes = [self.nodes[0], self.nmid, self.nodes[-1]]

    def __eq__(self, other):
        """Compare this face to other face. True if node lists equal.

        Args:
            other (Face): the other face we are comparing

        Returns:
            bool: True if face node lists are equal, False otherwise.
        """
        if self.nodes == other.nodes:
            return True
        else:
            return False

    def __str__(self):
        """Returns string listing object type, id number and nodes."""
        node1 = self.nodes[0].id
        node2 = self.nodes[1].id
        fid = self.id
        eid = self.element.id
        tup = (fid, eid, node1, node2)
        res = 'Face %i on element %i with nodes [%i,%i]' % tup
        return res


class Node(object):
    """Makes a mesh node.

    Args:
        node_num (int): node number
        x (float): x-coordinate
        y (float): y-coordinate
        z (float): z-coordinate

    Attributes:
        id (int): node number
        x (float): x-coordinate
        y (float): y-coordinate
        z (float): z-coordinate
        order (int): 1 or 2, 1 if it is a corner node, 2 if it is a midside node
        elements (set): a set of elements that contain this node
        faces (set): a set of faces that contain this node
    """

    def __init__(self, node_num, x, y, z):
        self.id = node_num
        self.x = x
        self.y = y
        self.z = z
        self.order = 1
        self.elements = set()
        self.faces = set()

    def set_order(self, order):
        """Sets the node order, must be 1 or 2.

        Args:
            order (int): node order, 1 is corner node, 2 is midside node
        """
        self.order = order

    def add_element(self, element):
        """Adds an element to this node's set of elements.

        Args:
            element (Element): element to add to set self.elements
        """
        self.elements.add(element)

    def add_face(self, face):
        """Adds a face to this node's set of faces.

        Args:
            face (Face): face to add to set self.faces
        """
        self.faces.add(face)

    def label(self, ax):
        """Labels the node on the passed Matplotlib axis."""
        ax.annotate(self.get_name(), (self.y, self.x))

    def __hash__(self):
        """Returns the node id as the hash."""
        return self.id

    def __eq__(self, other):
        """Returns boolean of id equality to other Node.

        Args:
            other (Node): the node to compare to this node.

        Returns:
            boolean: True if ids are equal, False if they are not
        """
        if isinstance(other, Node):
            return self.id == other.id
        else:
            return False

    def ccx(self):
        """Returns a string defining the node for a ccx inp file."""
        val = '%i, %f, %f, %f' % (self.id, self.x, self.y, self.z)
        return val

    def get_name(self):
        """Returns the string node name."""
        name = 'N%i' % (self.id)
        return name

    def __str__(self):
        """Returns string listing object type, id number and (x,y) coords."""
        val = 'Node, id %i, (x,y)=(%f,%f)' % (self.id, self.x, self.y)
        return val



"""This module defines the Component class, which is used to make components.

Components are collections that loads are applied to.
"""
from . import base_classes
from . import geometry

class Component(base_classes.Idobj):
    """Makes a component to store loads, boundary conditions, matl, thickness.

    A component must store a list of items where the type is identical for
    all items.

    Args:
        item_list (list): list of items to store
        ctype (str): type of component. Options:

            - 'nodes': all element nodes (corner and mid-side)
            - 'n1': element corner nodes
            - 'faces': faces
            - 'elements': elements

        cname (str): component name
    """

    def __init__(self, item_list, ctype, cname=''):
        self.items = item_list
        self.ctype = ctype
        self.name = cname+'_'+ctype
        base_classes.Idobj.__init__(self)

    def __eq__(self, other):
        """ Returns bool telling if self == other."""
        if isinstance(other, Component):
            if self.items == other.items and self.ctype == other.ctype:
                return True
            return False
        else:
            return False

    def __hash__(self):
        """Returns a hash of self instance."""
        return self.id

    def set_name(self, name):
        """Set the component name to the passed name.

        Args:
          name (str): passed name, component name is changed to this
        """
        self.name = name

    def get_name(self):
        """Returns the component name created with its id number."""
        return 'C'+str(self.id)

    def get_children(self):
        """Returns a list the child ctype items of the current component."""
        res = []
        for i in self.items:
            children = getattr(i, self.ctype)
            if isinstance(children, list):
                res += children
            else:
                res += [children]
        return res

    def ccx(self):
        """Writes a component for Calculix ccx solver.

        Returns a list strings, where each item is a line to write
        to a Calculix .inp text file.

        Note:
            Only node and element components should use this.
        """
        res = []
        items_per_line = 6
        firstline = ''
        if self.ctype == 'nodes':
            # node component, displacement
            firstline = '*NSET,NSET='+self.name
        elif self.ctype == 'elements':
            # element component, materials
            firstline = '*ELSET,ELSET='+self.name
        elif self.ctype == 'faces':
            # face component, contact surface
            firstline = '*SURFACE,NAME=%s,TYPE=ELEMENT' % self.name            
        res.append(firstline)
        items = self.get_children()
        if self.ctype == 'faces':
            for face in items:
                enum = face.element.id
                fnum = face.id
                res.append("%i,S%i" % (enum, fnum))
        elif self.ctype in ['nodes', 'elements']:
            grouped_items = base_classes.chunk_list(items, items_per_line)
            for group in grouped_items:
                item_ids = [str(x.id) for x in group]
                line = ', '.join(item_ids)
                if group != grouped_items[-1]:
                    line += ','
                res.append(line)
        return res

    def write_cgx(self):
        """Writes a component for Calculix cgx prerocessor.

        Returns a list strings, where each item is a line to write
        to a Calculix .fbd text file.

        Note:
            Only line and point node components should use this.
        """
        res = []

        # set the type of component
        parent = 'l'
        if (isinstance(self.items[0], geometry.Line) or
                isinstance(self.items[0], geometry.Arc)):
            parent = 'l'
        elif isinstance(self.items[0], geometry.Point):
            parent = 'p'

        # pull the list of entities
        alist = ' '.join([a.get_name() for a in self.items])

        if self.ctype == 'n1':
            # write component of line element first order nodes
            res.append('seta '+self.name+' '+parent+' '+alist)
        elif self.ctype == 'nodes':
            res.append('seta '+self.name+' '+parent+' '+alist)
            res.append('comp '+self.name+' do')
            # write component of all line element nodes
        elif self.ctype == 'f':
            # write component of line element faces
            res.append('seta '+self.name+' '+parent+' '+alist)
            res.append('comp '+self.name+' do')
            res.append('comp '+self.name+' do')
        return res

    def write_gmsh(self):
        """Writes a component for gmsh mesher.

        Returns a list strings, where each item is a line to write
        to a Calculix .geo text file.

        Note:
            Only line and point node components should use this.
        """
        res = []

        # set the type of component
        parent = 'l'
        if (isinstance(self.items[0], geometry.Line) or
                isinstance(self.items[0], geometry.Arc)):
            parent = 'l'
        elif isinstance(self.items[0], geometry.Point):
            parent = 'p'

        # pull the list of entities
        alist = ','.join([str(a.id) for a in self.items])

        line = ''
        if parent == 'l':
            line = "Physical Line('"+self.name+"') = {" + alist + '};'
        elif parent == 'p':
            line = "Physical Point('"+self.name+"') = {" + alist + '};'
        res.append(line)
        return res



from setuptools import setup, find_packages
with open('pycalculix/version.py') as f: exec(f.read())

setup(
    name = "pycalculix",
    install_requires = ['matplotlib >= 1.3.1', 'numpy', 'dxfgrabber >= 0.8.0', 'requests', 'pytest'],
    python_requires='>=3',
    version = __version__,
    description = "Python 3 library to build and solve finite element analysis (FEA) models in Calculix.",
    author = "Justin Black",
    author_email = "justin.a.black@gmail.com",
    packages = find_packages(),
    include_package_data=True,
    url = "http://justinablack.com/pycalculix/",
    entry_points = {
        'console_scripts': ['pycalculix-add-feaprograms=pycalculix.installer:add',
                            'pycalculix-remove-feaprograms=pycalculix.installer:remove'],
    },
    keywords = ["FEA", "Finite Element Analysis", "Calculix", "Mechanical Engineering", "CAD"],
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 7 - Inactive",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "License :: OSI Approved :: Apache Software License",
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering"],
    long_description = """\
'''
Python 3 library to automate and build finite element analysis (FEA) models in Calculix.
----------------------------------------------------------------------------------------
Meshing uses Calculix or GMSH.
Website: http://justinablack.com/pycalculix/
Source Code: https://github.com/spacether/pycalculix
Documentation: https://pythonhosted.org/pycalculix/

Useful applications of Pycalculix:
-Trade studies for plane stress, plane strain, or axisymmetric parts
-Quick Kt analysis of 2D geometry
-Learning finite element analyis (FEA) and Python

Notes:
I build a chunker in python which tries to cut big areas (> 5 sides) which
cgx can't mesh into smaller areas (<= 5 sides) which are meshable in cgx.
The chunker may not always be able to cut areas correctly.

Elements Supported:
Axisymmetric, plane stress, and plane strain elements are supported.
First and second order triangles and quadrilaterals are supported.
  First order elements only have corner nodes
  Second order elements have corner and mid-side nodes
Second order elements produce more accurate results
Setting element divisions on lines is supported

Geometry Building:
One can build separate parts made of points, lines, arcs, and areas.
One can draw a part made of straight lines, then smooth out corners by adding
blends/fillets with the part method: part.fillet_lines(L1, L2, arc_radius)
The new fillet will be tangent to both adjacent lines.

Loading:
Force loading
Constant pressure
Linearly varying pressure (water pressure)
Gravity
Rotational speed forces
Displacement constraints are supported
Loads are stored on geometry primitives (points lines etc) and can be applied
before or after meshing.

Examples:
https://github.com/spacether/pycalculix/tree/master/examples
"""
)
